"""Logging utility.
"""

import atexit
import functools
import logging
import os
import sys
from time import localtime, strftime
from typing import Union

from termcolor import colored

from dosma.utils import env

__all__ = ["setup_logger"]


class _ColorfulFormatter(logging.Formatter):
    """
    This class is adapted from Facebook's detectron2:
    https://github.com/facebookresearch/detectron2/blob/master/detectron2/utils/logger.py
    """

    def __init__(self, *args, **kwargs):
        self._root_name = kwargs.pop("root_name") + "."
        self._abbrev_name = kwargs.pop("abbrev_name", "")
        if len(self._abbrev_name):
            self._abbrev_name = self._abbrev_name + "."
        super(_ColorfulFormatter, self).__init__(*args, **kwargs)

    def formatMessage(self, record):
        record.name = record.name.replace(self._root_name, self._abbrev_name)
        log = super(_ColorfulFormatter, self).formatMessage(record)
        if record.levelno == logging.WARNING:
            prefix = colored("WARNING", "red", attrs=["blink"])
        elif record.levelno == logging.ERROR or record.levelno == logging.CRITICAL:
            prefix = colored("ERROR", "red", attrs=["blink", "underline"])
        else:
            return log
        return prefix + " " + log


@functools.lru_cache()  # so that calling setup_logger multiple times won't add many handlers
def setup_logger(
    output: Union[str, bool] = True,
    color=True,
    name="dosma",
    abbrev_name=None,
    stream_lvl=None,
    overwrite_handlers: bool = False,
):
    """Initialize the dosma logger.

    Args:
        output (str | bool): A file name or a directory to save log or a boolean.
            If ``True``, logs will save to the default dosma log location
            (:func:`dosma.utils.env.log_file_path`).
            If ``None`` or ``False``, logs will not be written to a file. This is not recommended.
            If a string and ends with ".txt" or ".log", assumed to be a file name.
            Otherwise, logs will be saved to `output/log.txt`.
        color (bool): If ``True``, logs printed to terminal (stdout) will be in color.
        name (str): The root module name of this logger.
        abbrev_name (str): An abbreviation of the module, to avoid long names in logs.
            Set to "" to not log the root module in logs.
            By default, will abbreviate "dosma" to "dm" and leave other
            modules unchanged.
        stream_lvl (int): The level for logging to console. Defaults to ``logging.DEBUG``
            if :func:`dosma.utils.env.debug()` is ``True``, else defaults to ``logging.INFO``.
        overwrite_handlers (bool): It ``True`` and logger with name ``name`` has logging handlers,
            these handlers will be removed before adding the new handlers. This is useful
            when to avoid having too many handlers for a logger.

    Returns:
        logging.Logger: A logger.

    Note:
        This method removes existing handlers from the logger.

    Examples:
        >>> setup_logger()  # how initializing logger is done most of the time
        >>> setup_logger("/path/to/save/dosma.log")  # save log to particular file
        >>> setup_logger(
        ... stream_lvl=logging.WARNING,
        ... overwrite_handlers=True)  # only prints warnings to console
    """
    if stream_lvl is None:
        stream_lvl = logging.DEBUG if env.debug() else logging.INFO

    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    logger.propagate = False

    # Clear handlers if they exist.
    is_new_logger = not logger.hasHandlers()
    if not is_new_logger and overwrite_handlers:
        logger.handlers.clear()

    if abbrev_name is None:
        abbrev_name = "dm" if name == "dosma" else name

    plain_formatter = logging.Formatter(
        "[%(asctime)s] %(name)s %(levelname)s: %(message)s", datefmt="%m/%d %H:%M:%S"
    )

    ch = logging.StreamHandler(stream=sys.stdout)
    ch.setLevel(stream_lvl)
    if color:
        formatter = _ColorfulFormatter(
            colored("[%(asctime)s %(name)s]: ", "green") + "%(message)s",
            datefmt="%m/%d %H:%M:%S",
            root_name=name,
            abbrev_name=str(abbrev_name),
        )
    else:
        formatter = plain_formatter
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    if output is not None and output is not False:
        if output is True:
            output = env.log_file_path()

        if output.endswith(".txt") or output.endswith(".log"):
            filename = output
        else:
            filename = os.path.join(output, "dosma.log")
        os.makedirs(os.path.dirname(filename), exist_ok=True)

        fh = logging.StreamHandler(_cached_log_stream(filename))
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(plain_formatter)
        logger.addHandler(fh)

    if is_new_logger and name == "dosma":
        logger.debug("\n" * 4)
        logger.debug("==" * 40)
        logger.debug("Timestamp: %s" % strftime("%Y-%m-%d %H:%M:%S", localtime()))
        logger.debug("\n\n")

    return logger


# cache the opened file object, so that different calls to `setup_logger`
# with the same file name can safely write to the same file.
@functools.lru_cache(maxsize=None)
def _cached_log_stream(filename):
    # use 1K buffer if writing to cloud storage
    io = open(filename, "a", buffering=1024 if "://" in filename else -1)
    atexit.register(io.close)
    return io



import logging
import os
from typing import Sequence

import numpy as np

from dosma import file_constants as fc
from dosma.core import quant_vals as qv
from dosma.core.fitting import MonoExponentialFit
from dosma.core.io import format_io_utils as fio_utils
from dosma.core.io.nifti_io import NiftiReader
from dosma.core.med_volume import MedicalVolume
from dosma.core.registration import apply_warp, register
from dosma.scan_sequences.scans import NonTargetSequence
from dosma.tissues.tissue import Tissue
from dosma.utils.cmd_line_utils import ActionWrapper

__all__ = ["CubeQuant"]

__EXPECTED_NUM_SPIN_LOCK_TIMES__ = 4
__R_SQUARED_THRESHOLD__ = 0.9
__INITIAL_T1_RHO_VAL__ = 70.0

__T1_RHO_LOWER_BOUND__ = 0.0
__T1_RHO_UPPER_BOUND__ = 500.0
__T1_RHO_DECIMAL_PRECISION__ = 3

_logger = logging.getLogger(__name__)


class CubeQuant(NonTargetSequence):
    """CubeQuant MRI sequence.

    Cubequant is a 3D fast-spin-echo (FSE) :math:`T_{1\\rho}`-weighted sequence.
    Acquisitions between spin-locks are susceptible to motion, and as a result,
    volumes within the scan have to be registered to each other (i.e. intra-registered).
    Intra-registration across different spin-locks is done by default upon construction.

    Moreover, CubeQuant scans often have lower resolution to increase SNR in practice.
    Because of the low-resolution, these scans are often registered to higher resolution
    target scans. This can be done using :meth:`CubeQuant.interregister`.
    """

    NAME = "cubequant"

    def __init__(self, volumes: Sequence[MedicalVolume], spin_lock_times: Sequence[float] = None):
        super().__init__(volumes=volumes)

        if spin_lock_times is None:
            try:
                if all(x.headers() is not None for x in self.volumes):
                    spin_lock_times = [x.get_metadata("EchoTime", float) for x in self.volumes]
            except (KeyError, AttributeError, RuntimeError) as e:
                raise ValueError(
                    f"Could not extract spin lock times from header. "
                    f"Please specify `spin_lock_times` argument - {e}"
                )
        self.spin_lock_times = spin_lock_times

    def intraregister(self):
        """Intra-register volumes.

        Patient could have moved between acquisition of different volumes,
        so different volumes of CubeQuant scan have to be registered with each other.

        The first spin lock time has the highest SNR, so it is used as the target.
        Volumes corresponding to the other spin lock times are registered to the target.

        Affine registration is done using Elastix.
        """
        self.__intraregister__()

    def interregister(self, target_path: str, target_mask_path: str = None):
        volumes = self.volumes
        spin_lock_times = self.spin_lock_times
        idxs = np.argsort(spin_lock_times)

        spin_lock_times = [spin_lock_times[i] for i in idxs]
        volumes = [volumes[i] for i in idxs]
        nr = NiftiReader()
        out_path = os.path.join(self.temp_path, "interregistered")
        os.makedirs(out_path, exist_ok=True)

        # TODO: Make these into parameters
        num_threads = 2
        num_workers = 0
        verbose = True

        base_image = volumes[0]
        moving = volumes[1:]

        if verbose:
            _logger.info("")
            _logger.info("==" * 40)
            _logger.info("Interregistering...")
            _logger.info("Target: {}".format(target_path))
            if target_mask_path is not None:
                _logger.info("Mask: {}".format(target_mask_path))
            _logger.info("==" * 40)

        if not target_mask_path:
            parameter_files = [fc.ELASTIX_RIGID_PARAMS_FILE, fc.ELASTIX_AFFINE_PARAMS_FILE]
            use_mask = None
        else:
            target_mask_path = self.__dilate_mask__(target_mask_path, out_path)
            parameter_files = [
                fc.ELASTIX_RIGID_INTERREGISTER_PARAMS_FILE,
                fc.ELASTIX_AFFINE_INTERREGISTER_PARAMS_FILE,
            ]
            use_mask = [False, True]

        out_reg, _ = register(
            target_path,
            base_image,
            parameters=parameter_files,
            output_path=out_path,
            sequential=True,
            collate=True,
            num_workers=num_workers,
            num_threads=num_threads,
            return_volumes=False,
            target_mask=target_mask_path,
            use_mask=use_mask,
            rtype=tuple,
            show_pbar=True,
        )
        out_reg = out_reg[0]

        reg_vols = [nr.load(out_reg.warped_file)]
        for mvg in moving:
            reg_vols.append(apply_warp(mvg, out_reg.transform))

        # Undo sorting by spin lock time.
        reverse_idxs = {v: i for i, v in enumerate(idxs)}
        reg_vols = [reg_vols[reverse_idxs[k]] for k in sorted(reverse_idxs.keys())]

        self.volumes = reg_vols

    def generate_t1_rho_map(self, tissue: Tissue, mask_path: str = None, num_workers: int = 0):
        """
        Generate 3D T1-rho map and r-squared fit map using mono-exponential
        fit across subvolumes acquired at different spin lock times.

        Args:
            tissue (Tissue): Tissue to generate quantitative value for.
            mask_path (:obj:`str`, optional): File path to mask of ROI to analyze.
                If specified, only voxels specified by mask will be fit.
                This can considerably speeds up computation.
            num_workers (int, optional): Number of subprocesses to use for fitting.
                If `0`, will execute on the main thread.

        Returns:
            qv.T1Rho: T1-rho fit for tissue.

        Raises:
            ValueError: If `mask_path` specifies non-binary volume.
        """
        spin_lock_times = self.spin_lock_times
        subvolumes_list = self.volumes

        # only calculate for focused region if a mask is available, this speeds up computation
        mask = tissue.get_mask()
        if mask_path is not None:
            mask = (
                fio_utils.generic_load(mask_path, expected_num_volumes=1)
                if isinstance(mask_path, (str, os.PathLike))
                else mask_path
            )

        mef = MonoExponentialFit(
            bounds=(__T1_RHO_LOWER_BOUND__, __T1_RHO_UPPER_BOUND__),
            tc0="polyfit",
            decimal_precision=__T1_RHO_DECIMAL_PRECISION__,
            num_workers=num_workers,
            verbose=True,
        )

        t1rho_map, r2 = mef.fit(spin_lock_times, subvolumes_list, mask=mask)

        quant_val_map = qv.T1Rho(t1rho_map)
        quant_val_map.add_additional_volume("r2", r2)

        tissue.add_quantitative_value(quant_val_map)

        return quant_val_map

    def __intraregister__(self):
        """Intra-register volumes.

        Patient could have moved between acquisition of different volumes,
        so different volumes of CubeQuant scan have to be registered with each other.

        The first spin lock time has the highest SNR, so it is used as the target.
        Volumes corresponding to the other spin lock times are registered to the target.

        Affine registration is done using Elastix.
        """
        volumes = self.volumes
        spin_lock_times = self.spin_lock_times
        idxs = np.argsort(spin_lock_times)

        spin_lock_times = [spin_lock_times[i] for i in idxs]
        volumes = [volumes[i] for i in idxs]
        out_path = os.path.join(self.temp_path, "interregistered")
        os.makedirs(out_path, exist_ok=True)

        # TODO: Make these into parameters
        num_threads = 2
        num_workers = 0
        verbose = True

        if verbose:
            _logger.info("")
            _logger.info("==" * 40)
            _logger.info("Intraregistering...")
            _logger.info("==" * 40)

        out_path = os.path.join(self.temp_path, "intraregister")
        _, reg_vols = register(
            volumes[0],
            volumes[1:],
            fc.ELASTIX_AFFINE_PARAMS_FILE,
            out_path,
            num_workers=num_workers,
            num_threads=num_threads,
            return_volumes=True,
            rtype=tuple,
            show_pbar=verbose,
        )
        reg_vols = [volumes[0]] + list(reg_vols)

        # Transfer header information
        reg_vols = [
            reg._partial_clone(volume=False, headers=vol.headers())
            for (reg, vol) in zip(reg_vols, volumes)
        ]

        # Undo sorting by spin lock time.
        reverse_idxs = {v: i for i, v in enumerate(idxs)}
        reg_vols = [reg_vols[reverse_idxs[k]] for k in sorted(reverse_idxs.keys())]

        self.volumes = reg_vols

    def _save(self, metadata, save_dir: str, fname_fmt=None, **kwargs):
        default_fmt = {MedicalVolume: "echo-{}"}
        default_fmt.update(fname_fmt if fname_fmt else {})
        return super()._save(metadata, save_dir, fname_fmt=default_fmt, **kwargs)

    @classmethod
    def from_dict(cls, data, force: bool = False) -> "CubeQuant":
        interregistered_dirpath = None
        if "subvolumes" in data:
            interregistered_dirpath = os.path.dirname(data.pop("subvolumes")[0])
        scan: CubeQuant = super().from_dict(data, force=force)
        if interregistered_dirpath is not None:
            subvolumes = scan.__load_interregistered_files__(interregistered_dirpath)
            cls.volumes = [subvolumes[k] for k in sorted(subvolumes.keys())]

        return scan

    @classmethod
    def cmd_line_actions(cls):
        """
        Provide command line information (such as name, help strings, etc)
        as list of dictionary.
        """
        intraregister_action = ActionWrapper(
            name=cls.intraregister.__name__, help="register volumes within this scan"
        )
        interregister_action = ActionWrapper(
            name=cls.interregister.__name__,
            help="register to another scan",
            param_help={
                "target_path": "path to target image in nifti format (.nii.gz)",
                "target_mask_path": "path to target mask in nifti format (.nii.gz)",
            },
            alternative_param_names={
                "target_path": ["tp", "target"],
                "target_mask_path": ["tm", "target_mask"],
            },
        )
        generate_t1rho_map_action = ActionWrapper(
            name=cls.generate_t1_rho_map.__name__,
            help="generate T1-rho map",
            aliases=["t1_rho"],
            param_help={
                "mask_path": "Mask used for fitting select voxels - in nifti format (.nii.gz)"
            },
        )

        return [
            (cls.intraregister, intraregister_action),
            (cls.interregister, interregister_action),
            (cls.generate_t1_rho_map, generate_t1rho_map_action),
        ]



from dosma.scan_sequences import mri

from dosma.scan_sequences import scans

from dosma.scan_sequences.mri import *  # noqa
from dosma.scan_sequences.scans import *  # noqa

__all__ = []
__all__.extend(mri.__all__)
__all__.extend(scans.__all__)



import numpy as np
import scipy.ndimage as sni
from scipy import optimize

from dosma.core.device import get_array_module
from dosma.utils import env

__all__ = ["circle_fit", "cart2pol"]


def circle_fit(x: np.ndarray, y: np.ndarray):
    """Fit a circle given (x, y) scatter points in a plane.

    Args:
      x (np.ndarray): x-coordinates of scatter points.
      y (np.ndarray): y-coordinates of scatter points.

    Returns:
        tuple[float]: Coordinates and radius of circle (center x, center y, radius).
    """
    ###
    # this function fit a circle given (x,y) scatter points in a plane.
    #
    # INPUT:
    #
    #   x................numpy array (n,) where n is the length of the array
    #   y................numpy array (n,) where n is the length of the array
    #
    # OUTPUT:
    #
    #   xc_2.............scalar, it is the coordinate x of the fitted circle
    #   yc_2.............scalar, it is the coordinate y of the fitted circle
    #   R_2..............scalar, it is the radius of the fitted circle
    ###

    # initialize the coordinate for the fitting procedure
    x_m = np.mean(x)
    y_m = np.mean(y)

    def calc_R(xc, yc):
        """
        Calculate the distance of each 2D points from the center (xc, yc).

        Args:
            xc: Center x.
            yc: Center y.
        """
        return np.sqrt((x - xc) ** 2 + (y - yc) ** 2)

    def f_2(c):
        """
        Calculate the algebraic distance between the 2D points
        and the mean circle centered at :math:`c=(xc, yc)`.

        Args:
            c (float): Circle center.
        """
        Ri = calc_R(*c)
        return Ri - Ri.mean()

    center_estimate = x_m, y_m
    center_2, ier = optimize.leastsq(f_2, center_estimate)

    xc_2, yc_2 = center_2
    Ri_2 = calc_R(xc_2, yc_2)
    R_2 = Ri_2.mean()
    # residu_2   = sum((Ri_2 - R_2)**2)
    # residu2_2  = sum((Ri_2**2-R_2**2)**2)

    return xc_2, yc_2, R_2


def cart2pol(x, y):
    """Convert cartesian coordinates to polar coordinates.

    Args:
        x: x-coordinate.
        y: y-coordinate.

    Returns:
        tuple[float, float]: radius (rho) and angle (phi).
    """
    rho = np.sqrt(x ** 2 + y ** 2)
    phi = np.arctan2(y, x)

    phi = phi * (180 / np.pi)  # degrees
    phi[phi == 180] = -180

    return rho, phi


def center_of_mass(input, labels=None, index=None):
    """
    Calculate the center of mass of the values of an array at labels.

    Args:
        input : ndarray
            Data from which to calculate center-of-mass. The masses can either
            be positive or negative.
        labels : ndarray, optional
            Labels for objects in `input`, as generated by `ndimage.label`.
            Only used with `index`. Dimensions must be the same as `input`.
        index : int or sequence of ints, optional
            Labels for which to calculate centers-of-mass. If not specified,
            all labels greater than zero are used. Only used with `labels`.

    Returns
    -------
    center_of_mass : tuple, or list of tuples
        Coordinates of centers-of-mass.

    Note:
        This is adapted from scipy.ndimage to support cupy.
    """
    _sni = sni
    if env.cupy_available():
        import cupy as cp

        if get_array_module(input) == cp:
            import cupyx.scipy.ndimage as csni

            _sni = csni

    return _sni.center_of_mass(input, labels=labels, index=index)



import logging
import sys
import tkinter as tk


class WidgetLogger(logging.StreamHandler):
    def __init__(self, widget):
        logging.Handler.__init__(self, sys.stdout)
        self.setLevel(logging.INFO)
        self.widget = widget
        self.widget.config(state="disabled")

    def emit(self, record):
        self.widget.config(state="normal")
        # Append message (record) to the widget
        self.widget.insert(tk.END, self.format(record) + "\n")
        self.widget.see(tk.END)  # Scroll to the bottom
        self.widget.config(state="disabled")



import os
import time

import scipy.io as sio

from utils import dicom_utils, io_utils

dataset_map = {
    "rep_02": ["005", "008"],
    "rep_03": ["005", "009"],
    "rep_04": ["005", "009"],
    "rep_05": ["005", "007"],
    "rep_06": ["005", "009"],
    "rep_07": ["005", "010"],
}
knee_sides = ["LEFT", "RIGHT"]

outLoc = io_utils.check_dir("/bmrNAS/people/arjun/dess_data/dess_bilateral_data/mat_files/")

dicomDir = "/bmrNAS/people/barma7/bilateral/dess_study/data/marco_study_copy/%s/%s/%s/"
segFileBase = "/bmrNAS/people/barma7/bilateral/dess_study/data/marco_study_copy/%s/data/%s/%s/fc/fc_manual.nii.gz"  # noqa

start_time = time.time()
counter = 0
for subject in dataset_map.keys():
    for scan_id in dataset_map[subject]:
        for knee in knee_sides:
            volume, _ = dicom_utils.load_dicom(dicomDir % (subject, scan_id, knee))
            volume = volume.volume
            echo1 = volume[..., 0::2]
            echo2 = volume[..., 1::2]

            seg = io_utils.load_nifti(segFileBase % (subject, scan_id, knee))
            seg = seg.volume

            save_dir = io_utils.check_dir(os.path.join(outLoc, subject, scan_id, knee))
            sio.savemat(
                os.path.join(save_dir, "matFile.mat"), {"echo1": echo1, "echo2": echo2, "seg": seg}
            )

    counter += 1
    print("Finished subject %d/%d" % (counter, len(dataset_map.keys())))

print("Time Elapsed: %0.2f seconds" % (time.time() - start_time))



from dosma.core.io import dicom_io, format_io_utils, nifti_io  # noqa: F401

from dosma.core.io.dicom_io import *  # noqa
from dosma.core.io.format_io import ImageDataFormat  # noqa
from dosma.core.io.format_io_utils import *  # noqa
from dosma.core.io.nifti_io import *  # noqa

__all__ = []
__all__.extend(dicom_io.__all__)
__all__.extend(["ImageDataFormat"])
__all__.extend(format_io_utils.__all__)
__all__.extend(nifti_io.__all__)



import inspect
import os
import warnings
from abc import ABC
from pathlib import Path
from typing import Any, Dict, Optional, Sequence, Set, Union

import pydicom

from dosma.core.io import format_io_utils as fio_utils
from dosma.core.io.dicom_io import DicomReader
from dosma.core.io.format_io import ImageDataFormat
from dosma.core.med_volume import MedicalVolume
from dosma.defaults import preferences
from dosma.tissues.tissue import Tissue
from dosma.utils import io_utils


def _contains_type(value, types):
    """Returns ``True`` if any value is an instance of ``types``."""
    if isinstance(value, types):
        return True
    if not isinstance(value, str) and isinstance(value, (Sequence, Set)) and len(value) > 0:
        return any(_contains_type(x, types) for x in value)
    elif isinstance(value, Dict):
        return _contains_type(value.keys(), types) or _contains_type(value.values(), types)
    return isinstance(value, types)


class ScanIOMixin(ABC):
    # This is just a summary on variables used in this abstract class,
    # the proper values/initialization should be done in child class.
    NAME: str
    __DEFAULT_SPLIT_BY__: Optional[str]
    _from_file_args: Dict[str, Any]

    @classmethod
    def from_dicom(
        cls,
        dir_or_files,
        group_by=None,
        ignore_ext: bool = False,
        num_workers: int = 0,
        verbose: bool = False,
        **kwargs,
    ):
        """Load scan from dicom files.

        Args:
            dir_or_files (str): The path to dicom directory or files.
            group_by: DICOM field tag name or tag number used to group dicoms. Defaults
                to scan's ``__DEFAULT_SPLIT_BY__``.
            ignore_ext (bool, optional): If `True`, ignore extension (`.dcm`)
                when loading dicoms from directory.
            num_workers (int, optional): Number of workers to use for loading.
            verbose (bool, optional): If ``True``, enable verbose logging for dicom loading.
            kwargs: Other keywords required to construct scan.

        Returns:
            The scan.
        """
        dr = DicomReader(num_workers, verbose)
        if group_by is None:
            group_by = cls.__DEFAULT_SPLIT_BY__
        volumes = dr.load(dir_or_files, group_by, ignore_ext)

        if isinstance(dir_or_files, (str, Path, os.PathLike)):
            dir_or_files = os.path.abspath(dir_or_files)
        else:
            dir_or_files = type(dir_or_files)([os.path.abspath(x) for x in dir_or_files])

        scan = cls(volumes, **kwargs)
        scan._from_file_args = {
            "dir_or_files": dir_or_files,
            "ignore_ext": ignore_ext,
            "group_by": group_by,
            "_type": "dicom",
        }

        return scan

    @classmethod
    def from_dict(cls, data: Dict[str, Any], force: bool = False):
        """Loads class from data dictionary.

        Args:
            data (Dict): The data.
            force (bool, optional): If ``True``, writes attributes even if they do not exist.
                Use with caution.

        Returns:
            The scan

        Examples:
            >>> scan = ... # some scan
            >>> filepath = scan.save("/path/to/base/directory")
            >>> scan_from_saved = type(scan).from_dict(io_utils.load_pik(filepath))
            >>> scan_from_dict = type(scan).from_dict(scan.__dict__)
        """
        # TODO: Add check for deprecated and converted attribute names.
        data = cls._convert_attr_name(data)

        # TODO: Convert metadata to appropriate type.
        # Converting metadata type is important when loading MedicalVolume data (for example).
        # The data is stored as a path, but should be loaded as a MedicalVolume.
        data = cls.load_custom_data(data)

        signature = inspect.signature(cls)
        init_metadata = {k: v for k, v in data.items() if k in signature.parameters}
        scan = cls(**init_metadata)
        for k in init_metadata.keys():
            data.pop(k)

        for k, v in data.items():
            if not hasattr(scan, k) and not force:
                warnings.warn(f"{cls.__name__} does not have attribute {k}. Skipping...")
                continue
            scan.__setattr__(k, v)

        return scan

    def save(
        self,
        path: str,
        save_custom: bool = False,
        image_data_format: ImageDataFormat = None,
        num_workers: int = 0,
    ):
        """Saves scan data to disk with option for custom saving.

        Custom saving may be useful to reduce redundant saving and/or save data in standard
        compatible formats (e.g. medical images - nifti/dicom), which are not feasible with
        python serialization libraries, like pickle.

        When ``save_custom=True``, this method overloads standard pickling with customizable
        saving by first saving data in customizable way (e.g. MedicalVolume -> Nifti file),
        and then pickling the reference to the saved object (e.g. Nifti filepath).

        Currently certain custom saving of objects such as ``pydicom.FileDataset`` and
        :cls:`Tissue` objects are not supported.

        To load the data, do the following:

        >>> filepath = scan.save("/path/to/directory", save_custom=True)
        >>> scan_loaded = type(scan).load(io_utils.load_pik(filepath))

        Args:
            path (str): Directory where data is stored.
            data_format (ImageDataFormat, optional): Format to save data.
                Defaults to ``preferences.image_data_format``.
            save_custom (bool, optional): If ``True``, saves data in custom way specified
                by :meth:`save_custom_data` in format specified by ``data_format``. For
                example, for default classes this will save :cls:`MedicalVolume` data
                to nifti/dicom files as specified by ``image_data_format``.
            image_data_format (ImageDataFormat, optional): The data format to save
                :cls:`MedicalVolume` data. Only used if save_custom is ``True``.
            num_workers (int, bool): Number of workers for saving custom data.
                Only used if save_custom is ``True``.

        Returns:
            str: The path to the pickled file.
        """
        if image_data_format is None:
            image_data_format = preferences.image_data_format

        save_dirpath = path  # self._save_dir(path)
        os.makedirs(save_dirpath, exist_ok=True)
        filepath = os.path.join(save_dirpath, "%s.data" % self.NAME)

        metadata: Dict = {}
        for attr in self.__serializable_variables__():
            metadata[attr] = self.__getattribute__(attr)

        if save_custom:
            metadata = self._save(
                metadata, save_dirpath, image_data_format=image_data_format, num_workers=num_workers
            )

        io_utils.save_pik(filepath, metadata)
        return filepath

    @classmethod
    def load(cls, path_or_data: Union[str, Dict], num_workers: int = 0):
        """Load scan.

        This method overloads the :func:`from_dict` method by supporting loading from a file
        in addition to the data dictionary. If loading and constructing a scan using
        :func:`from_dict` fails, defaults to loading data from original dicoms
        (if ``self._from_file_args`` is initialized).

        Args:
            path_or_data (Union[str, Dict]): Pickle file to load or data dictionary.
            num_workers (int, optional): Number of workers to use for loading.

        Returns:
            ScanSequence: Of type ``cls``.

        Raises:
            ValueError: If ``scan`` cannot be constructed.
        """
        if isinstance(path_or_data, (str, Path, os.PathLike)):
            if os.path.isdir(path_or_data):
                path_or_data = os.path.join(path_or_data, f"{cls.NAME}.data")

            if not os.path.isfile(path_or_data):
                raise FileNotFoundError(f"File {path_or_data} does not exist")
            data = io_utils.load_pik(path_or_data)
        else:
            data = path_or_data

        try:
            scan = cls.from_dict(data)
            return scan
        except Exception:
            warnings.warn(
                f"Failed to load {cls.__name__} from data. Trying to load from dicom file."
            )

        data = cls._convert_attr_name(data)
        data = cls.load_custom_data(data, num_workers=num_workers)

        scan = None
        if "_from_file_args" in data:
            dicom_args = data.pop("_from_file_args")
            assert dicom_args.pop("_type") == "dicom"
            scan = cls.from_dicom(**dicom_args, num_workers=num_workers)
        elif "dicom_path" in data:
            # Backwards compatibility
            dicom_path = data.pop("dicom_path")
            ignore_ext = data.pop("ignore_ext", False)
            group_by = data.pop("split_by", cls.__DEFAULT_SPLIT_BY__)
            scan = cls.from_dicom(
                dicom_path, ignore_ext=ignore_ext, group_by=group_by, num_workers=num_workers
            )

        if scan is None:
            raise ValueError(f"Data is insufficient to construct {cls.__name__}")

        for k, v in data.items():
            if not hasattr(scan, k):
                warnings.warn(f"{cls.__name__} does not have attribute {k}. Skipping...")
                continue
            scan.__setattr__(k, v)

        return scan

    def save_data(
        self, base_save_dirpath: str, data_format: ImageDataFormat = preferences.image_data_format
    ):
        """Deprecated: Alias for :func:`self.save`."""
        warnings.warn(
            "save_data is deprecated since v0.0.13 and will no longer be "
            "available in v0.1. Use `save` instead.",
            DeprecationWarning,
        )
        return self.save(base_save_dirpath, data_format)

    def _save(
        self,
        metadata: Dict[str, Any],
        save_dir: str,
        fname_fmt: Dict[Union[str, type], str] = None,
        **kwargs,
    ):
        if fname_fmt is None:
            fname_fmt = {}

        default_fname_fmt = {MedicalVolume: "image-{}"}
        for k, v in default_fname_fmt.items():
            if k not in fname_fmt:
                fname_fmt[k] = v

        for attr in metadata.keys():
            val = metadata[attr]
            path = fname_fmt.get(attr, None)

            if path is None:
                path = os.path.abspath(os.path.join(save_dir, attr))
            if not os.path.isabs(path):
                path = os.path.join(save_dir, attr, path)
            try:
                metadata[attr] = self.save_custom_data(val, path, fname_fmt, **kwargs)
            except Exception as e:
                raise RuntimeError(f"Failed to save metadata {attr} - {e}")

        return metadata

    def save_custom_data(
        self, metadata, paths, fname_fmt: Dict[Union[str, type], str] = None, **kwargs
    ):
        """
        Finds all attributes of type MedicalVolume or Sequence/Mapping to MedicalVolume
        and saves them.
        """
        if isinstance(metadata, (Dict, Sequence, Set)):
            if isinstance(paths, str):
                paths = [paths] * len(metadata)
            else:
                assert len(paths) == len(metadata)

        if isinstance(metadata, Dict):
            keys = metadata.keys()
            if isinstance(paths, Dict):
                paths = [paths[k] for k in keys]
            paths = [os.path.join(_path, f"{k}") for k, _path in zip(keys, paths)]
            values = self.save_custom_data(metadata.values(), paths, fname_fmt, **kwargs)
            metadata = {k: v for k, v in zip(keys, values)}
        elif not isinstance(metadata, str) and isinstance(metadata, (Sequence, Set)):
            values = list(metadata)
            paths = [os.path.join(_path, "{:03d}".format(i)) for i, _path in enumerate(paths)]
            values = [
                self.save_custom_data(_x, _path, fname_fmt, **kwargs)
                for _x, _path in zip(values, paths)
            ]
            if not isinstance(values, type(metadata)):
                metadata = type(metadata)(values)
            else:
                metadata = values
        else:
            formatter = [fname_fmt.get(x) for x in type(metadata).__mro__]
            formatter = [x for x in formatter if x is not None]
            if len(formatter) == 0:
                formatter = None
            else:
                formatter = formatter[0]
            metadata = self._save_custom_data_base(metadata, paths, formatter, **kwargs)

        return metadata

    def _save_custom_data_base(self, metadata, path, formatter: str = None, **kwargs):
        """The base condition for :meth:`save_custom_data`.

        Args:
            metadata (Any): The data to save.
            path (str): The path to save the data.
            formatter (str, optional): If provided, this formatted string
                will be used to format ``path``.
        """
        out = {"__dtype__": type(metadata)}
        # TODO: Add support for num workers.
        # num_workers = kwargs.pop("num_workers", 0)

        if formatter:
            path = os.path.join(os.path.dirname(path), formatter.format(os.path.basename(path)))

        if isinstance(metadata, MedicalVolume):
            image_data_format = kwargs.get("image_data_format", preferences.image_data_format)
            # TODO: Once, `files` property added to MedicalVolume, check if property is
            # set before doing saving.
            path = fio_utils.convert_image_data_format(path, image_data_format)
            metadata.save_volume(path, data_format=image_data_format)
            out["__value__"] = path
        else:
            out = metadata

        return metadata

    @classmethod
    def _convert_attr_name(cls, data: Dict[str, Any]):
        return data

    @classmethod
    def load_custom_data(cls, data: Any, **kwargs):
        """Recursively converts data to appropriate types.

        By default, this loads all :class:`MedicalVolume` objects from their corresponding paths.

        Args:
            data (Any): The data. Can either be the dictionary or the metadata value.
                If the data corresponds to a custom type, it should have the following
                schema:

                {
                    '__dtype__': The type of the data.
                    '__value__': The value from which object of type __dtype__ can be constructed.
                }

            **kwargs: Keyword Arguments to pass to :meth:`_load_custom_data_base`.

        Returns:
            Any: The loaded metadata.
        """
        dtype = type(data)
        if isinstance(data, Dict) and "__value__" in data:
            dtype = data["__dtype__"]
            data = data["__value__"]

        if issubclass(dtype, Dict):
            keys = cls.load_custom_data(data.keys(), **kwargs)
            values = cls.load_custom_data(data.values(), **kwargs)
            data = {k: v for k, v in zip(keys, values)}
        elif not issubclass(dtype, str) and issubclass(dtype, (list, tuple, set)):
            data = dtype([cls.load_custom_data(x, **kwargs) for x in data])
        else:
            data = cls._load_custom_data_base(data, dtype, **kwargs)

        return data

    @classmethod
    def _load_custom_data_base(cls, data, dtype=None, **kwargs):
        """The base condition for :meth:`load_custom_data`.

        Args:
            data:
            dtype (type): The data type.

        Return:
            The loaded data.
        """
        if dtype is None:
            dtype = type(data)

        # TODO: Add support for loading with num_workers
        num_workers = kwargs.get("num_workers", 0)
        if isinstance(data, str) and issubclass(dtype, MedicalVolume):
            data = fio_utils.generic_load(data, num_workers=num_workers)

        return data

    def __serializable_variables__(
        self, ignore_types=(pydicom.FileDataset, pydicom.Dataset, Tissue), ignore_attrs=()
    ) -> Set:
        """
        By default, all instance attributes are serialized except those
        corresponding to headers, :class:`MedicalVolume`(s), or :class:`Tissues`.
        Properties and class attributes are also not stored. Class attributes are
        indentified using the PEP8 nomenclature of all caps variables.

        Note:
            This method has not been profiled, but times may be large if
            the instance contains many variables. Currently this is not
            cached as attributes values can change and, as a result, must
            be inspected.
        """
        serializable = []
        for attr, value in self.__dict__.items():
            if attr in ignore_attrs or _contains_type(value, ignore_types):
                continue
            if attr.startswith("temp") or attr.startswith("_temp"):
                continue
            if attr.upper() == attr or (attr.startswith("__") and attr.endswith("__")):
                continue
            if callable(value) or isinstance(value, property):
                continue
            serializable.append(attr)

        return set(serializable)



"""Analysis for patellar cartilage.

Attributes:
    BOUNDS (dict): Upper bounds for quantitative values.
"""
import itertools
import os
import warnings

import numpy as np
import pandas as pd
import scipy.ndimage as sni

from dosma.core.device import get_array_module
from dosma.core.quant_vals import QuantitativeValueType
from dosma.defaults import preferences
from dosma.tissues.tissue import Tissue, largest_cc
from dosma.utils import io_utils

import matplotlib.pyplot as plt

# milliseconds
BOUNDS = {
    QuantitativeValueType.T2: 60.0,
    QuantitativeValueType.T1_RHO: 100.0,
    QuantitativeValueType.T2_STAR: 50.0,
}

__all__ = ["PatellarCartilage"]


class PatellarCartilage(Tissue):
    """Handles analysis and visualization for patellar cartilage."""

    ID = 3
    STR_ID = "pc"
    FULL_NAME = "patellar cartilage"

    # Expected quantitative values
    T1_EXPECTED = 1000  # milliseconds

    # Region Keys
    _ANTERIOR_KEY = 0
    _POSTERIOR_KEY = 1
    _CORONAL_KEYS = [_ANTERIOR_KEY, _POSTERIOR_KEY]

    _MEDIAL_KEY = 0
    _LATERAL_KEY = 1
    _SAGITTAL_KEYS = [_MEDIAL_KEY, _LATERAL_KEY]

    _REGION_DEEP_KEY = 0
    _REGION_SUPERFICIAL_KEY = 1
    _TOTAL_AXIAL_KEY = -1

    def __init__(self, weights_dir: str = None, medial_to_lateral: bool = None):
        super().__init__(weights_dir=weights_dir, medial_to_lateral=medial_to_lateral)

        self.regions_mask = None

    def unroll_coronal(self, quant_map: np.ndarray):
        """Unroll patellar cartilage in the coronal direction.

        Because patellar cartilage is flat, "unrolling" is projecting the patellar
        cartilage onto the coronal axis.

        Args:
            quant_map (np.ndarray):
        """
        mask = self.__mask__.volume

        assert (
            self.regions_mask is not None
        ), "region_mask not initialized. Should be initialized when mask is set"
        region_mask_deep_superficial = self.regions_mask[..., 0]

        superficial = (
            (region_mask_deep_superficial == self._REGION_SUPERFICIAL_KEY) * mask * quant_map
        )
        superficial[superficial == 0] = np.nan
        superficial = np.nanmean(superficial, axis=2)

        deep = (region_mask_deep_superficial == self._REGION_DEEP_KEY) * mask * quant_map
        deep[deep == 0] = np.nan
        deep = np.nanmean(deep, axis=2)

        total = mask * quant_map
        total[total == 0] = np.nan
        total = np.nanmean(total, axis=2)

        return total, superficial, deep

    def split_regions(self, base_map):
        """Split patellar cartilage into deep/superficial regions.

        For patellar cartilage, the superficial/deep transition occurs in
        the anterior/posterior (A/P) direction. The boundary is determined
        for each non-zero 1D column spanning independently by the local
        center-of-mass (COM). The medial/lateral (M/L) plane is computed
        using the global COM.

        Args:
            base_map (ndarray): Binary 3D mask with orientation (SI, AP, ML/LM).
                If `self.medial_to_lateral`, last dimension should be ML.
        """
        if np.sum(base_map) == 0:
            warnings.warn("No mask for `%s` was found." % self.FULL_NAME)

        # Superficial/Deep (A/P)
        locs = base_map.sum(axis=1).nonzero()
        voxels = base_map[locs[0], :, locs[1]]
        com_sup_inf = np.asarray(
            [
                int(np.ceil(sni.measurements.center_of_mass(voxels[i, :])[0]))
                for i in range(voxels.shape[0])
            ]
        )
        region_mask_sup_deep = np.full(base_map.shape, self._REGION_DEEP_KEY)
        for i in range(len(com_sup_inf)):
            region_mask_sup_deep[
                locs[0][i], : com_sup_inf[i], locs[1][i]
            ] = self._REGION_SUPERFICIAL_KEY

        # M/L
        cum_ml = np.nonzero(base_map.sum(axis=(0, 1)))[0]  # noqa: F841
        # midpoint_ml = int(np.ceil((np.min(cum_ml) + np.max(cum_ml)) / 2))
        midpoint_ml = int(np.ceil(sni.measurements.center_of_mass(base_map)[2]))
        region_mask_med_lat = np.full(base_map.shape, self._LATERAL_KEY)
        medial_span = slice(0, midpoint_ml) if self.medial_to_lateral else slice(midpoint_ml, None)
        region_mask_med_lat[:, :, medial_span] = self._MEDIAL_KEY

        self.regions_mask = np.stack([region_mask_sup_deep, region_mask_med_lat], axis=-1)

    def __calc_quant_vals__(self, quant_map, map_type):
        subject_pid = self.pid

        super().__calc_quant_vals__(quant_map, map_type)

        assert (
            self.regions_mask is not None
        ), "region_mask not initialized. Should be initialized when mask is set"

        quant_map_volume = quant_map.volume
        mask = self.__mask__.volume
        quant_map_volume = mask * quant_map_volume

        deep_superficial_map = self.regions_mask[..., 0]
        med_lat_map = self.regions_mask[..., 1]

        axial_names = ["superficial", "deep", "total"]
        sagittal_names = ["medial", "lateral"]

        pd_header = ["Subject", "Location", "Condyle", "Mean", "Std", "Median"]
        pd_list = []

        regions = itertools.product(
            [self._REGION_SUPERFICIAL_KEY, self._REGION_DEEP_KEY, self._TOTAL_AXIAL_KEY],
            [self._MEDIAL_KEY, self._LATERAL_KEY],
        )
        for axial, sagittal in regions:
            if axial == self._TOTAL_AXIAL_KEY:
                axial_map = np.asarray(
                    deep_superficial_map == self._REGION_SUPERFICIAL_KEY, dtype=np.float32
                ) + np.asarray(deep_superficial_map == self._REGION_DEEP_KEY, dtype=np.float32)
                axial_map = np.asarray(axial_map, dtype=np.bool)
            else:
                axial_map = deep_superficial_map == axial

            sagittal_map = med_lat_map == sagittal

            curr_region_mask = quant_map_volume * axial_map * sagittal_map
            curr_region_mask = curr_region_mask[curr_region_mask != 0]

            # discard all values that are 0
            c_mean = np.nanmean(curr_region_mask)
            c_std = np.nanstd(curr_region_mask)
            c_median = np.nanmedian(curr_region_mask)

            row_info = [
                subject_pid,
                axial_names[axial],
                sagittal_names[sagittal],
                c_mean,
                c_std,
                c_median,
            ]

            pd_list.append(row_info)

        # Generate 2D unrolled matrix
        total, superficial, deep = self.unroll_coronal(quant_map.volume)

        df = pd.DataFrame(pd_list, columns=pd_header)
        qv_name = map_type.name
        maps = [
            {
                "title": "%s superficial" % qv_name,
                "data": superficial,
                "xlabel": "Slice",
                "ylabel": "Angle (binned)",
                "filename": "%s_superficial" % qv_name,
                "raw_data_filename": "%s_superficial.data" % qv_name,
            },
            {
                "title": "%s deep" % qv_name,
                "data": deep,
                "xlabel": "Slice",
                "ylabel": "Angle (binned)",
                "filename": "%s_deep" % qv_name,
                "raw_data_filename": "%s_deep.data" % qv_name,
            },
            {
                "title": "%s total" % qv_name,
                "data": total,
                "xlabel": "Slice",
                "ylabel": "Angle (binned)",
                "filename": "%s_total" % qv_name,
                "raw_data_filename": "%s_total.data" % qv_name,
            },
        ]

        self.__store_quant_vals__(maps, df, map_type)

    def set_mask(self, mask, use_largest_cc: bool = True):
        xp = get_array_module(mask.A)
        if use_largest_cc:
            msk = xp.asarray(largest_cc(mask.A), dtype=xp.uint8)
        else:
            msk = xp.asarray(mask.A, dtype=xp.uint8)
        mask_copy = mask._partial_clone(volume=msk)
        super().set_mask(mask_copy)

        self.split_regions(self.__mask__.volume)

    def __save_quant_data__(self, dirpath):
        """Save quantitative data and 2D visualizations of patellar cartilage

        Check which quantitative values (T2, T1rho, etc) are defined for
        patellar cartilage and analyze these:

        1. Save 2D total, superficial, and deep visualization maps
        2. Save {'medial', 'lateral'}, {'anterior', 'posterior'},
            {'superior', 'inferior', 'total'} data to excel file

        :param dirpath: base filepath to save data
        """
        q_names = []
        dfs = []

        for quant_val in QuantitativeValueType:
            if quant_val.name not in self.quant_vals.keys():
                continue

            q_names.append(quant_val.name)
            q_val = self.quant_vals[quant_val.name]
            dfs.append(q_val[1])

            q_name_dirpath = io_utils.mkdirs(os.path.join(dirpath, quant_val.name.lower()))
            for q_map_data in q_val[0]:
                filepath = os.path.join(q_name_dirpath, q_map_data["filename"])
                xlabel = ""
                ylabel = ""
                title = q_map_data["title"]
                data_map = q_map_data["data"]

                axs_bounds = self.__get_axis_bounds__(data_map, leave_buffer=True)

                plt.clf()

                upper_bound = BOUNDS[quant_val]
                if preferences.visualization_use_vmax:
                    # Hard bounds - clipping
                    plt.imshow(data_map, cmap="jet", vmin=0.0, vmax=BOUNDS[quant_val])
                else:
                    # Try to use a soft bounds
                    if np.sum(data_map <= upper_bound) == 0:
                        plt.imshow(data_map, cmap="jet", vmin=0.0, vmax=BOUNDS[quant_val])
                    else:
                        warnings.warn(
                            "%s: Pixel value exceeded upper bound (%0.1f). Using normalized scale."
                            % (quant_val.name, upper_bound)
                        )
                        plt.imshow(data_map, cmap="jet")

                plt.xlabel(xlabel)
                plt.ylabel(ylabel)
                plt.title(title)
                plt.ylim(axs_bounds[0])
                plt.gca().invert_yaxis()
                plt.xlim(axs_bounds[1])
                # plt.axis('tight')
                clb = plt.colorbar()
                clb.ax.set_ylabel("(ms)")
                plt.savefig(filepath)

                # Save data
                raw_data_filepath = os.path.join(
                    q_name_dirpath, "raw_data", q_map_data["raw_data_filename"]
                )
                io_utils.save_pik(raw_data_filepath, data_map)

        if len(dfs) > 0:
            io_utils.save_tables(os.path.join(dirpath, "data.xlsx"), dfs, q_names)



from dosma.tissues import femoral_cartilage, meniscus, patellar_cartilage, tibial_cartilage
from dosma.tissues.femoral_cartilage import *  # noqa
from dosma.tissues.meniscus import *  # noqa
from dosma.tissues.patellar_cartilage import *  # noqa
from dosma.tissues.tibial_cartilage import *  # noqa
from dosma.tissues.tissue import *  # noqa

__all__ = []
__all__.extend(femoral_cartilage.__all__)
__all__.extend(meniscus.__all__)
__all__.extend(patellar_cartilage.__all__)
__all__.extend(tibial_cartilage.__all__)



import matplotlib

matplotlib.use("TkAgg")

from dosma.gui.ims import DosmaViewer  # noqa: E402
from dosma.utils.logger import setup_logger  # noqa: E402

# Initialize logger for the GUI.
setup_logger()

app = DosmaViewer()
app.mainloop()



import inspect
import tkinter as tk
from tkinter import IntVar
from typing import Dict

import Pmw

from dosma.cli import SEGMENTATION_MODEL_KEY, SEGMENTATION_WEIGHTS_DIR_KEY
from dosma.gui.gui_utils import gui_utils as gutils
from dosma.models import SUPPORTED_MODELS
from dosma.models.seg_model import SegModel
from dosma.tissues.tissue import Tissue


class ScanReader:
    def __init__(self, parent):
        self.parent = parent
        self.hbox = None
        self.action_box = None
        self.params: Dict = {}
        self.action_var = None
        self._action_bool = IntVar()
        self.balloon = None

    def load_scan(self, scan_class):
        self._action_bool = IntVar()
        if self.hbox:
            self.hbox.destroy()
        if self.action_box:
            self.action_box.destroy()
        self.action_var = None
        self.balloon = Pmw.Balloon()

        cmd_line_actions = scan_class.cmd_line_actions()
        hbox = tk.Frame(self.parent)
        hbox.pack(anchor="nw", side="top")

        buttons = []
        count = 0
        for a_method, a_description in cmd_line_actions:
            b = tk.Radiobutton(
                hbox,
                text=a_description.name,
                value=count,
                command=lambda v=(a_method, a_description): self.show_action_params(v[0], v[1]),
                variable=self._action_bool,
            )
            self.balloon.bind(b, a_description.help)
            buttons.append(b)
            count += 1

        self._action_bool.set(-1)

        for b in buttons:
            b.pack(anchor="nw", side="left", padx=5)

        self.hbox = hbox

    def show_action_params(self, action, action_wrapper):
        self.action_var = action_wrapper.name
        if self.action_box:
            self.action_box.destroy()

        if self.params:
            self.params = {}

        hbox = tk.Frame(self.parent)
        hbox.pack(anchor="nw", side="top")
        self.action_box = hbox

        func_signature = inspect.signature(action)
        parameters = func_signature.parameters

        for param_name in parameters.keys():
            param = parameters[param_name]
            param_type = param.annotation
            param_default = param.default

            if param_name == "self" or param_type is Tissue:
                continue

            # # see if the type is a custom type, if not handle it as a basic type
            is_custom_arg = param_type in CUSTOM_TYPE_TO_GUI
            if is_custom_arg:
                CUSTOM_TYPE_TO_GUI[param_type](self.params, hbox, self.balloon)
                continue

            param_var = gutils.convert_base_type_to_gui(
                param_name,
                param_type,
                param_default,
                hbox,
                balloon=self.balloon,
                param_help=action_wrapper.get_param_help(param_name),
            )

            # map parameter name --> variable, is_required
            # if you have a non zero default value, it must be specified.
            is_required = (param_type is not bool and param_default == inspect._empty) or (
                param_type in [float, int] and bool(param_default)
            )
            self.params[param_name] = (param_var, is_required)

    def get_cmd_line_str(self):
        if not self.action_var:
            raise ValueError("No action selected. Select an action to continue.")
        cmd_line_str = "%s" % self.action_var
        for param_name in self.params:
            param_var, add_arg = self.params[param_name]

            if add_arg and not param_var.get():
                raise ValueError('"%s" must have a value' % param_name)

            if param_var.get():
                cmd_line_str += " --%s" % param_name
                if add_arg:
                    cmd_line_str += " %s" % param_var.get()

        return cmd_line_str


def add_segmentation_gui_parser(params, hbox, balloon):
    # add model
    param_name, param_type, param_default = SEGMENTATION_MODEL_KEY, str, None
    param_var = gutils.convert_base_type_to_gui(
        param_name,
        param_type,
        param_default,
        hbox,
        balloon=balloon,
        param_help="segmentation models",
        options=SUPPORTED_MODELS,
    )
    params[param_name] = (param_var, param_type is not bool)

    # add weights directory
    param_name, param_type, param_default = SEGMENTATION_WEIGHTS_DIR_KEY, str, None
    param_var = gutils.convert_base_type_to_gui(
        param_name,
        param_type,
        param_default,
        hbox,
        balloon=balloon,
        param_help="path to weights directory",
    )
    params[param_name] = (param_var, param_type is not bool)


CUSTOM_TYPE_TO_GUI = {SegModel: add_segmentation_gui_parser}



"""
Command line interface declaration for knee-related analyses

@author: Arjun Desai
        (C) Stanford University, 2019
"""

import logging
import uuid

from dosma.core.quant_vals import QuantitativeValueType as QV
from dosma.defaults import preferences
from dosma.tissues import FemoralCartilage, Meniscus, PatellarCartilage, TibialCartilage

KNEE_KEY = "knee"
MEDIAL_TO_LATERAL_KEY = "ml"
TISSUES_KEY = "tissues"
LOAD_KEY = "load"
SAVE_KEY = "save"
PID_KEY = "pid"

SUPPORTED_TISSUES = [FemoralCartilage, Meniscus, TibialCartilage, PatellarCartilage]
SUPPORTED_QUANTITATIVE_VALUES = [QV.T2, QV.T1_RHO, QV.T2_STAR]

_logger = logging.getLogger(__name__)


def knee_parser(base_parser):
    """Parse command line input related to knee

    :param base_parser: the base parser to add knee subcommand to
    """
    parser_tissue = base_parser.add_parser(
        KNEE_KEY, help="calculate/analyze quantitative data for knee"
    )

    parser_tissue.add_argument(
        "--%s" % MEDIAL_TO_LATERAL_KEY,
        action="store_const",
        const=True,
        default=False,
        help="defines slices in sagittal direction going from medial -> lateral",
    )

    parser_tissue.add_argument(
        "--%s" % PID_KEY, nargs="?", default=str(uuid.uuid4()), help="specify pid"
    )

    for tissue in SUPPORTED_TISSUES:
        parser_tissue.add_argument(
            "--%s" % tissue.STR_ID,
            action="store_const",
            default=False,
            const=True,
            help="analyze %s" % tissue.FULL_NAME,
        )

    qvs_dict = {}
    for qv in SUPPORTED_QUANTITATIVE_VALUES:
        qv_name = qv.name.lower()
        qvs_dict[qv_name] = qv
        parser_tissue.add_argument(
            "--%s" % qv_name,
            action="store_const",
            const=True,
            default=False,
            help="quantify %s" % qv_name,
        )

    parser_tissue.set_defaults(func=handle_knee)


def handle_knee(vargin):
    """Handle parsing command-line input for knee subcommand
    :param vargin:
    :return:
    """
    tissues = vargin[TISSUES_KEY]
    load_path = vargin[LOAD_KEY]
    medial_to_lateral = vargin[MEDIAL_TO_LATERAL_KEY]
    pid = vargin[PID_KEY]

    if tissues is None or len(tissues) == 0:
        _logger.info("Computing for all supported knee tissues...")
        tissues = []
        for t in SUPPORTED_TISSUES:
            tissues.append(t())

    # Get all supported quantitative values
    qvs = []
    for qv in SUPPORTED_QUANTITATIVE_VALUES:
        if vargin[qv.name.lower()]:
            qvs.append(qv)

    if len(qvs) == 0:
        _logger.info("Computing for all supported quantitative values...")
        qvs = SUPPORTED_QUANTITATIVE_VALUES

    for tissue in tissues:
        tissue.pid = pid
        tissue.medial_to_lateral = medial_to_lateral
        tissue.load_data(load_path)

        _logger.info("")
        _logger.info("==" * 40)
        _logger.info(tissue.FULL_NAME)
        _logger.info("==" * 40)

        for qv in qvs:
            # load file
            _logger.info("Analyzing %s" % qv.name.lower())
            tissue.calc_quant_vals()

    for tissue in tissues:
        tissue.save_data(vargin[SAVE_KEY], data_format=preferences.image_data_format)

    return tissues



from dosma.core import io

from dosma.core import (
    device,
    fitting,
    med_volume,
    numpy_routines,
    orientation,
    quant_vals,
    registration,
)

from dosma.core.device import *  # noqa
from dosma.core.fitting import *  # noqa
from dosma.core.io import *  # noqa
from dosma.core.med_volume import *  # noqa
from dosma.core.orientation import *  # noqa
from dosma.core.registration import *  # noqa

__all__ = ["numpy_routines", "quant_vals"]
__all__.extend(device.__all__)
__all__.extend(fitting.__all__)
__all__.extend(io.__all__)
__all__.extend(med_volume.__all__)
__all__.extend(orientation.__all__)
__all__.extend(registration.__all__)



"""Utils for data I/O.
"""

import os
from pathlib import Path
from typing import List, Union

from dosma.core.io.dicom_io import DicomReader, DicomWriter
from dosma.core.io.format_io import DataReader, DataWriter, ImageDataFormat
from dosma.core.io.nifti_io import NiftiReader, NiftiWriter
from dosma.core.med_volume import MedicalVolume

__all__ = [
    "read",
    "write",
    "get_reader",
    "get_writer",
    "get_filepath_variations",
    "convert_image_data_format",
    "generic_load",
]

_READERS = {ImageDataFormat.dicom: DicomReader, ImageDataFormat.nifti: NiftiReader}
_WRITERS = {ImageDataFormat.dicom: DicomWriter, ImageDataFormat.nifti: NiftiWriter}


def get_reader(data_format: ImageDataFormat) -> DataReader:
    """Return a DataReader corresponding to the given data format.

    Args:
        data_format (ImageDataFormat): Data format to read.

    Returns:
        DataReader: Reader for given format.
    """
    return _READERS[data_format]()


def get_writer(data_format: ImageDataFormat) -> DataWriter:
    """Return a DataWriter corresponding to given data format.

    Args:
        data_format (ImageDataFormat): Data format to write.

    Returns:
        DataWriter: Writer for given format.
    """
    return _WRITERS[data_format]()


def convert_image_data_format(
    file_or_dir_path: Union[str, Path, os.PathLike], new_data_format: ImageDataFormat
) -> str:
    """Change a file or directory name to convention of another data format.

    Args:
        file_or_dir_path (str): File or directory path.
        new_data_format (ImageDataFormat): Data format convention for file/directory name.

    Returns:
        str: File/directory path based on convention of new data format.

    Raises:
        NotImplementedError: If conversion from current image data format
            to new image data format not found.
    """
    current_format = ImageDataFormat.get_image_data_format(file_or_dir_path)

    if current_format == new_data_format:
        return file_or_dir_path

    if current_format == ImageDataFormat.dicom and new_data_format == ImageDataFormat.nifti:
        dirname = os.path.dirname(file_or_dir_path)
        basename = os.path.basename(file_or_dir_path)
        return os.path.join(dirname, "{}.nii.gz".format(basename))

    if current_format == ImageDataFormat.nifti and new_data_format == ImageDataFormat.dicom:
        dirname = os.path.dirname(file_or_dir_path)
        basename = os.path.basename(file_or_dir_path)
        basename = basename.split(".", 1)[0]
        return os.path.join(dirname, basename)

    raise NotImplementedError(
        "{} -> {} not implemented".format(current_format.name, new_data_format.name)
    )


def get_filepath_variations(file_or_dir_path: Union[str, Path, os.PathLike]):
    """Get file paths using convention for all different image data formats.

    Args:
        file_or_dir_path (str): File or directory path.

    Returns:
        list[str]: File paths corresponding to naming conventions of different ImageDataFormats.
    """
    filepath_variations = []
    for io_format in ImageDataFormat:
        filepath_variations.append(convert_image_data_format(file_or_dir_path, io_format))
    return filepath_variations


def generic_load(file_or_dir_path: Union[str, Path, os.PathLike], expected_num_volumes: int = None):
    """Load MedicalVolume(s) from a file or directory without knowing data format.

    Args:
        file_or_dir_path (str): File or directory path.
        expected_num_volumes (int, `optional`): Number of volumes expected.
            If specified, assert if number of loaded
            volumes != expected num volumes. Defaults to `None`.

    Returns:
        `list[MedicalVolume]` or `MedicalVolume`: Volume(s) loaded.
            If ``expected_num_volumes = 1``, returns :class:``MedicalVolume``.

    Raises:
        ValueError: If multiple file paths corresponding to different ImageDataFormats exist.
        FileNotFoundError: If file path or corresponding versions of file path not found.
    """
    possible_filepaths = get_filepath_variations(file_or_dir_path)
    exist_path = None

    for fp in possible_filepaths:
        if os.path.exists(fp):
            if exist_path is not None:
                raise ValueError(
                    "Ambiguous loading state - multiple possible files to load from %s"
                    % str(possible_filepaths)
                )
            exist_path = fp

    if exist_path is None:
        raise FileNotFoundError(
            "No file associated with basename %s found" % os.path.basename(file_or_dir_path)
        )

    io_format = ImageDataFormat.get_image_data_format(exist_path)
    r = get_reader(io_format)
    vols = r.load(exist_path)

    if expected_num_volumes is None:
        return vols

    if type(vols) is not list:
        vols = [vols]

    assert len(vols) == expected_num_volumes, "Expected %d volumes, got %d" % (
        expected_num_volumes,
        len(vols),
    )

    if len(vols) == 1:
        return vols[0]

    return vols


def read(
    path: Union[str, Path, os.PathLike],
    data_format: ImageDataFormat = None,
    unpack: bool = False,
    **kwargs
) -> Union[MedicalVolume, List[MedicalVolume]]:
    """Read MedicalVolume(s) from file.

    Args:
        path (str): File/directory path.
        data_format (ImageDataFormat, optional): Data format
            (e.g. ``'dicom'``, ``'nifti'``, etc.). Use this is disambiguate between different
            data formats. If this function is not working, try using this argument.
            If not provided, dosma will try to infer data format from file extension.
        unpack (bool, optional): If ``True`` and only 1 volume is loaded, return a single
            volume instead of a list of volumes. This only applied to dicom loading.
        **kwargs: Additional keyword arguments passed to the data format reader.

    Returns:
        MedicalVolume | List[MedicalVolume]: Volume(s) loaded.

    Examples:
        >>> dm.read("/path/to/multi-echo/dicom/folder", group_by="EchoNumbers")
        >>> dm.read("/path/to/ct/dicom/folder")
        >>> dm.read("/path/to/ct/nifti/file.nii.gz", mmap=True)
    """
    if data_format is None:
        data_format = ImageDataFormat.get_image_data_format(path)
    elif isinstance(data_format, str):
        data_format = ImageDataFormat[data_format]

    out = get_reader(data_format).load(path, **kwargs)
    if unpack and isinstance(out, (tuple, list)) and len(out) == 1:
        out = out[0]
    return out


def write(
    vol: MedicalVolume,
    path: Union[str, Path, os.PathLike],
    data_format: ImageDataFormat = None,
    **kwargs
) -> None:
    """Write MedicalVolume to file.

    Args:
        vol (MedicalVolume): Volume to write.
        path (str): File/directory path.
        data_format (ImageDataFormat, optional): Data format
            (e.g. ``'dicom'``, ``'nifti'``, etc.). Use this is disambiguate between different
            data formats. If this function is not working, try using this argument.
            If not provided, dosma will try to infer data format from file extension.
        **kwargs: Additional keyword arguments passed to the data format writer.

    Raises:
        FileNotFoundError: If file path or corresponding versions of file path not found.
    """
    if data_format is None:
        data_format = ImageDataFormat.get_image_data_format(path)
    elif isinstance(data_format, str):
        data_format = ImageDataFormat[data_format]

    get_writer(data_format).save(vol, path, **kwargs)



"""
Functions for loading Keras models

@author: Arjun Desai
        (C) Stanford University, 2019
"""
import os
import yaml
from functools import partial
from typing import Sequence

from dosma.models.oaiunet2d import IWOAIOAIUnet2D, IWOAIOAIUnet2DNormalized, OAIUnet2D
from dosma.models.seg_model import SegModel

__all__ = ["get_model", "SUPPORTED_MODELS"]

# Network architectures currently supported
__SUPPORTED_MODELS__ = [OAIUnet2D, IWOAIOAIUnet2D, IWOAIOAIUnet2DNormalized]

# Initialize supported models for the command line
SUPPORTED_MODELS = [x.ALIASES[0] for x in __SUPPORTED_MODELS__]


def get_model(model_str, input_shape, weights_path, **kwargs):
    """Get a Keras model
    :param model_str: model identifier
    :param input_shape: tuple or list of tuples for initializing input(s) into Keras model
    :param weights_path: filepath to weights used to initialize Keras model
    :return: a Keras model
    """
    for m in __SUPPORTED_MODELS__:
        if model_str in m.ALIASES or model_str == m.__name__:
            return m(input_shape, weights_path, **kwargs)

    raise LookupError("%s model type not supported" % model_str)


def model_from_config(cfg_file_or_dict, weights_dir=None, **kwargs) -> SegModel:
    """Builds a new model from a config file.

    This function is useful for building models that have similar structure/architecture
    to existing models supported in DOSMA, but have different weights and categories.
    The config defines what dosma model should be used as a base, what weights should be loaded,
    and what are the categories.

    The config file should be a yaml file that has the following keys:
        * "DOSMA_MODEL": The base model that exists in DOSMA off of which data should be built.
        * "CATEGORIES": The categories that are supposed to be loaded.
        * "WEIGHTS_FILE": The basename of (or full path to) weights that should be loaded.

    Args:
        cfg_file_or_dict (str or dict): The yaml file or dictionary corresponding to the config.
        weights_dir (str): The directory where weights are stored. If not specified, assumes
            "WEIGHTS_FILE" field in the config is the full path to the weights.
        **kwargs: Keyword arguments for base model `__init__`

    Returns:
        SegModel: A segmentation model with appropriate changes to `generate_mask` to produce
            the right masks.
    """

    def _gen_mask(func, *_args, **_kwargs):
        out = func(*_args, **_kwargs)
        if isinstance(out, dict):
            # Assumes that the dict is ordered, which it is for python>=3.6
            out = out.values()
        elif not isinstance(out, Sequence):
            out = [out]
        if not len(categories) == len(out):
            raise ValueError("Got {} outputs, but {} categories".format(len(out), len(categories)))
        return {cat: out for cat, out in zip(categories, out)}

    if isinstance(cfg_file_or_dict, str):
        with open(cfg_file_or_dict, "r") as f:
            cfg = yaml.load(f)
    else:
        cfg = cfg_file_or_dict

    base_model = cfg["DOSMA_MODEL"]
    categories = cfg["CATEGORIES"]
    weights = cfg["WEIGHTS_FILE"]
    if not os.path.isfile(weights):
        assert weights_dir, "`weights_dir` must be specified"
        weights = os.path.join(weights_dir, cfg["WEIGHTS_FILE"])

    try:
        model: SegModel = get_model(base_model, weights_path=weights, force_weights=True, **kwargs)
    except LookupError as e:
        raise LookupError("BASE_MODEL '{}' not supported \n{}".format(base_model, e))

    # Override generate mask function
    model.generate_mask = partial(_gen_mask, model.generate_mask)

    return model



import tkinter as tk
from tkinter import ttk
from typing import Dict

import Pmw

from dosma.cli import GPU_KEY
from dosma.core.io.format_io import ImageDataFormat
from dosma.defaults import preferences
from dosma.utils import env

if env.package_available("tensorflow"):
    from tensorflow.python.client import device_lib
else:
    device_lib = None


CUDA_DEVICES_STR = "CUDA_VISIBLE_DEVICES"
SUPPORTED_IMAGE_DATA_FORMATS = list(ImageDataFormat)
LARGE_FONT = ("Verdana", 18)


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


TYPE_CAST = {bool: tk.BooleanVar, str: tk.StringVar, int: tk.IntVar, float: tk.DoubleVar}


class CommandLineFlagGUI:
    def __init__(self, cmd_line_metadata: dict, **kwargs):
        self._cmd_line_metadata = cmd_line_metadata
        self._name = cmd_line_metadata["name"]
        self._default = cmd_line_metadata["default"]
        self._help = cmd_line_metadata["help"]
        self._type = cmd_line_metadata["type"]
        self._choices = (
            cmd_line_metadata["choices"] if "choices" in cmd_line_metadata.keys() else None
        )

        # Tkinter Variable used for binding with GUI.
        self._tk_var = TYPE_CAST[self._type]()
        self._tk_var.set(self._default)

        # GUI formatting.
        self._label_format = kwargs.get("label_format") if "label_format" in kwargs else "%s: "

        # Memory
        self._widget = None

    def draw(self, frame: tk.Frame, **kwargs):
        """Draw element best suited for command line metadata.
        :param frame: The frame where to add GUI element
        :type frame: A Tkinter frame
        """
        draw_cmd = {
            str: lambda root: self._draw_str_gui(root),
            bool: lambda root: self._draw_bool_gui(root),
            int: lambda root: self._draw_number_gui(root, "float"),
            float: lambda root: self._draw_number_gui(root, "float"),
        }

        padx = kwargs.get("padx") if "padx" in kwargs else 5
        balloon = kwargs.get("balloon") if "balloon" in kwargs else None

        hbox = tk.Frame(frame)
        hbox.pack(side="top", anchor="nw")

        label = tk.Label(hbox, text=self._label_format % self._name)
        label.pack(side="left", anchor="nw", padx=padx)

        if self._choices:
            t = self._draw_list_gui(hbox)
        else:
            t = draw_cmd[self._type](hbox)

        t.pack(side="left", anchor="nw", padx=padx)
        self._widget = t

        if balloon:
            balloon.bind(label, self._help)

        return hbox

    def _draw_str_gui(self, root):
        return tk.Entry(root, textvariable=self.tk_var)

    def _draw_bool_gui(self, root):
        return tk.Checkbutton(root, variable=self.tk_var)

    def _draw_number_gui(self, root, dtype="float"):
        vcmd = (
            root.register(self._validate_number),
            dtype,
            "%d",
            "%i",
            "%P",
            "%s",
            "%S",
            "%v",
            "%V",
            "%W",
        )
        return tk.Entry(root, textvariable=self.tk_var, validate="all", validatecommand=vcmd)

    def _draw_int_gui(self, root):
        vcmd = (
            root.register(self._validate_number),
            "float",
            "%d",
            "%i",
            "%P",
            "%s",
            "%S",
            "%v",
            "%V",
            "%W",
        )
        return tk.Entry(root, textvariable=self.tk_var, validate="all", validatecommand=vcmd)

    def _validate_number(
        self,
        dtype,
        action,
        index,
        value_if_allowed,
        prior_value,
        text,
        validation_type,
        trigger_type,
        widget_name,
    ):
        if trigger_type not in ["key", "focusout"]:
            return True

        if not value_if_allowed:
            if trigger_type == "key":
                return True
            else:
                self.tk_var.set(self._default)
                self._widget["validate"] = validation_type
                return False

        try:
            eval(dtype)(value_if_allowed)
            return True
        except ValueError:
            return False

    def _draw_list_gui(self, root):
        options = self._choices
        return tk.OptionMenu(root, self.tk_var, *options)

    @property
    def tk_var(self):
        return self._tk_var


class PreferencesManager(metaclass=Singleton):
    def __init__(self):
        self.frame = None
        self.balloon = None

        self.gui_manager: Dict = {}
        self.gui_elements: Dict = {}

        self._init_gpu_preferences()

        # Init preferences
        self._preference_elements = {}
        preferences_metadata = preferences.cmd_line_flags()
        for preference in preferences_metadata.keys():
            self._preference_elements[preference] = CommandLineFlagGUI(
                preferences_metadata[preference], label_format="%s:\t"
            )

    def _init_gpu_preferences(self):
        gpu_vars = []
        if device_lib is not None:
            local_device_protos = device_lib.list_local_devices()
            for x in local_device_protos:
                if x.device_type == "GPU":
                    bool_var = tk.BooleanVar()
                    x_id = x.name.split(":")[-1]
                    gpu_vars.append((x_id, bool_var))

        self.gui_manager["gpu"] = gpu_vars

    @property
    def gpus(self) -> str:
        if device_lib is None:
            return None

        gpu_ids = []
        local_device_protos = device_lib.list_local_devices()
        for x in local_device_protos:
            if x.device_type == "GPU":
                x_id = x.name.split(":")[-1]
                gpu_ids.append(x_id)

        if len(gpu_ids) == 0:
            return None

        gpu_str = ""
        for ind, var in enumerate(self.gui_manager["gpu"]):
            if not var.get():
                continue
            gpu_str += "%s," % gpu_ids[ind]

        if len(gpu_str) == 0:
            return None

        gpu_str = gpu_str[:-1]  # remove last comma
        return gpu_str

    def _restore_preference_default(self):
        for preference in self._preference_elements.keys():
            self._preference_elements[preference].tk_var.set(preferences.get(preference))

    def __display_gui(self):
        self._restore_preference_default()
        self.balloon = Pmw.Balloon()

        _label = tk.Label(self.frame, text="Preferences", font=LARGE_FONT)
        _label.pack()

        hboxes = []
        # show gpu options
        if self.gui_manager["gpu"]:
            f = tk.Frame(self.frame)
            gpu_checkboxes = []
            gpu_label = tk.Label(f, text="GPU:")
            self.balloon.bind(gpu_label, "Select gpus to use for analysis")
            gpu_label.pack()
            for x_id, bool_var in self.gui_manager["gpu"]:
                c = tk.Checkbutton(f, text=x_id, variable=bool_var)
                c.pack()
                gpu_checkboxes.append(c)
            hboxes.append(f)

        # Add command line governed preferences
        for preference in self._preference_elements.keys():
            hbox = self._preference_elements[preference].draw(self.frame, balloon=self.balloon)
            hboxes.append(hbox)

        for f in hboxes:
            f.pack(side="top", anchor="w", pady=10)

        # Add apply settings and save preferences buttons
        # apply settings: save preferences for this session only
        # save settings: changes preferences in file
        hbox = tk.Frame(self.frame)
        apply_settings_button = ttk.Button(
            hbox, text="Apply Settings", command=lambda: self._apply_settings()
        )
        self.balloon.bind(apply_settings_button, "Apply settings for this session")
        save_settings_button = ttk.Button(
            hbox, text="Save Settings", command=lambda: self._save_settings()
        )
        self.balloon.bind(save_settings_button, "Save settings for all future sessions")
        apply_settings_button.pack(side="left", anchor="nw", padx=5)
        save_settings_button.pack(side="left", anchor="nw", padx=5)
        hbox.pack(side="bottom", anchor="se")

    def _apply_settings(self):
        for preference in self._preference_elements.keys():
            new_val = self._preference_elements[preference].tk_var.get()
            preferences.set(preference, new_val)

    def _save_settings(self):
        self._apply_settings()
        preferences.save()

    def show_window(self, parent):
        window = tk.Toplevel(parent)
        self.frame = window
        self.__display_gui()

    def get_cmd_line_str(self):
        gpus = self.gpus
        cmd_line_str = ""
        if gpus:
            cmd_line_str += "--%s %s " % (GPU_KEY, gpus)

        return cmd_line_str.strip()



import os
import shutil
import unittest

import numpy as np
import pandas as pd

from dosma.utils import io_utils

from .. import util

IO_UTILS_DATA = os.path.join(util.UNITTEST_DATA_PATH, "io_utils")


class UtilsTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        io_utils.mkdirs(IO_UTILS_DATA)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(IO_UTILS_DATA)

    def test_h5(self):
        filepath = os.path.join(IO_UTILS_DATA, "sample.h5")
        datas = [{"type": np.random.rand(10, 45, 2), "type2": np.random.rand(13, 95, 4)}]

        for data in datas:
            io_utils.save_h5(filepath, data)
            data2 = io_utils.load_h5(filepath)

            assert len(list(data.keys())) == len(list(data2.keys()))

            for key in data.keys():
                assert (data[key] == data2[key]).all()

    def test_pik(self):
        filepath = os.path.join(IO_UTILS_DATA, "sample.pik")
        datas = {"type": np.random.rand(10, 45, 2), "type2": np.random.rand(13, 95, 4)}

        io_utils.save_pik(filepath, datas)
        datas2 = io_utils.load_pik(filepath)

        for data in datas:
            assert (datas[data] == datas2[data]).all()

    def test_save_tables(self):
        df = pd.DataFrame({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
        path = os.path.join(IO_UTILS_DATA, "table.xlsx")
        io_utils.save_tables(path, [df])

        df2 = pd.read_excel(path, engine="openpyxl")
        assert np.all(df == df2)



from dosma.scan_sequences.mri import cones, cube_quant, mapss, qdess  # noqa: F401

from dosma.scan_sequences.mri.cones import *  # noqa
from dosma.scan_sequences.mri.cube_quant import *  # noqa
from dosma.scan_sequences.mri.mapss import *  # noqa
from dosma.scan_sequences.mri.qdess import *  # noqa

__all__ = []
__all__.extend(cones.__all__)
__all__.extend(cube_quant.__all__)
__all__.extend(mapss.__all__)
__all__.extend(qdess.__all__)



import os
import warnings

import numpy as np
import pandas as pd
import scipy.ndimage as sni

from dosma.core.device import get_array_module
from dosma.core.io.format_io import ImageDataFormat
from dosma.core.med_volume import MedicalVolume
from dosma.core.quant_vals import QuantitativeValueType
from dosma.defaults import preferences
from dosma.tissues.tissue import Tissue, largest_cc
from dosma.utils import img_utils, io_utils
from dosma.utils.geometry_utils import cart2pol, circle_fit

import matplotlib.pyplot as plt

# milliseconds
BOUNDS = {
    QuantitativeValueType.T2: 80.0,
    QuantitativeValueType.T1_RHO: 100.0,
    QuantitativeValueType.T2_STAR: 80.0,
}

__all__ = ["FemoralCartilage"]


class FemoralCartilage(Tissue):
    """Handles analysis and visualization for femoral cartilage.

    This class extends functionality from `Tissue`.

    For visualization, the femoral cartilage is unrolled onto a 2D plane using angular binning [1].

    References:
        [1] Monu UD, Jordan CD, Samuelson BL, Hargreaves BA, Gold GE, McWalter EJ.
        Cluster analysis of quantitative MRI T2 and :math:`T1\\rho` relaxation times of
        cartilage identifies differences between healthy and ACL-injured individuals at 3T."
        Osteoarthritis and cartilage 2017;25(4):513-520.
    """

    ID = 1
    STR_ID = "fc"
    FULL_NAME = "femoral cartilage"

    # Expected quantitative values
    T1_EXPECTED = 1200  # milliseconds

    # Keys correspond to integer representing bit location for each region
    # bit string: 'T D S M L A C P' (stored as integer)
    # Coronal Keys
    _POSTERIOR_KEY = 2 ** 0
    _CENTRAL_KEY = 2 ** 1
    _ANTERIOR_KEY = 2 ** 2
    _CORONAL_KEYS = [_POSTERIOR_KEY, _CENTRAL_KEY, _ANTERIOR_KEY]

    # Sagittal Keys
    _MEDIAL_KEY = 2 ** 3
    _LATERAL_KEY = 2 ** 4
    _SAGITTAL_KEYS = [_MEDIAL_KEY, _LATERAL_KEY]

    # Axial Keys
    _DEEP_KEY = 2 ** 5
    _SUPERFICIAL_KEY = 2 ** 6
    _TOTAL_AXIAL_KEY = 2 ** 7
    _AXIAL_KEYS = [_DEEP_KEY, _SUPERFICIAL_KEY, _TOTAL_AXIAL_KEY]

    # Do not change order of below.
    # Order reflects order of _CORONAL_KEYS, _SAGITTAL_KEYS, _AXIAL_KEYS
    _AXIAL_NAMES = ["deep", "superficial", "total"]
    _SAGITTAL_NAMES = ["medial", "lateral"]
    _CORONAL_NAMES = ["posterior", "central", "anterior"]

    ML_BOUNDARY = None
    ACP_BOUNDARY = None

    def __init__(self, weights_dir=None, medial_to_lateral=None):
        super().__init__(weights_dir=weights_dir)

        self.regions_mask = None
        self.theta_bins = None

        self.medial_to_lateral = medial_to_lateral

    def split_regions(
        self, base_map: np.ndarray, thickness_divisor=0.5, num_bins=72, theta=(-270, 90)
    ):
        """Split volume into anatomical regions.

        Pixels corresponding to femoral cartilage are divided across 3 planes:
            - Coronal: Posterior, Central, or Anterior
            - Sagittal: Medial, Lateral
            - Axial: Deep, Superficial

        For example, a pixel could correspond to the Posterior Lateral Deep region of
            femoral cartilage.

        Args:
            base_map (np.ndarray): 3D numpy array typically corresponding to volume to split.

        Returns:
            np.ndarray: 4D numpy array (region, height, width, depth).
                Saved in variable ``self.regions``.
        """
        dtheta = 360 / num_bins
        theta_min, theta_max = tuple(theta)

        mask = self.__mask__.volume

        mask = mask * np.nan_to_num(base_map)

        height, width, num_slices = mask.shape

        # STEP 1: PROJECTING AND CYLINDRICAL FIT
        segmented_t2maps_projected = np.max(mask, 2)  # Project segmented T2maps on sagittal axis
        non_zero_element = np.nonzero(segmented_t2maps_projected)

        xc_fit, yc_fit, R_fit = circle_fit(
            non_zero_element[1], non_zero_element[0]
        )  # fit a circle to projected cartilage tissue

        # STEP 2: SLICE BY SLICE BINNING
        yv, xv = np.meshgrid(range(height), range(width), indexing="ij")

        rho, theta = cart2pol(xv - xc_fit, yc_fit - yv)
        theta = (theta >= 90) * (theta - 360) + (theta < 90) * theta  # range: [-270, 90)

        assert (np.min(theta) >= theta_min) and (
            np.max(theta) < theta_max
        ), "Expected Theta range is [{:d}, {:d}) degrees. Received min: {:d} max: {:d})".format(
            theta_min, theta_max, np.min(theta), np.max(theta)
        )

        theta_bins = np.floor((theta - theta_min) / dtheta)

        # STEP 3: COMPUTE THRESHOLD RADII
        # TODO: This step takes a long time
        rhos_threshold_volume = np.zeros(mask.shape)
        for curr_slice in range(num_slices):
            mask_slice = mask[..., curr_slice]

            for curr_bin in range(num_bins):
                rhos_valid = rho[np.logical_and(mask_slice > 0, theta_bins == curr_bin)]
                if len(rhos_valid) == 0:
                    continue

                rho_min = np.min(rhos_valid)
                rho_max = np.max(rhos_valid)

                rho_threshold = thickness_divisor * (rho_max - rho_min) + rho_min
                rhos_threshold_volume[theta_bins == curr_bin, curr_slice] = rho_threshold

        regions_volume = np.asarray(np.zeros(mask.shape), dtype=np.uint16)

        # anterior/central/posterior division
        # Central region occupies middle 30 degrees, anterior on left, posterior on right
        anterior_region = self._ANTERIOR_KEY * (theta < -105)
        central_region = self._CENTRAL_KEY * np.logical_and((theta >= -105), (theta < -75))
        posterior_region = self._POSTERIOR_KEY * (theta >= -75)
        acp_map = anterior_region + central_region + posterior_region
        acp_volume = np.asarray(np.stack([acp_map] * num_slices, axis=-1), dtype=np.uint16)
        regions_volume += acp_volume

        # medial/lateral division
        # take into account scanning direction
        center_of_mass = sni.measurements.center_of_mass(mask)
        com_slicewise = center_of_mass[-1]
        ml_volume = np.asarray(np.zeros(mask.shape), dtype=np.uint16)

        if self.medial_to_lateral:
            ml_volume[..., : int(np.ceil(com_slicewise))] = self._MEDIAL_KEY
            ml_volume[..., int(np.ceil(com_slicewise)) :] = self._LATERAL_KEY
        else:
            ml_volume[..., : int(np.ceil(com_slicewise))] = self._LATERAL_KEY
            ml_volume[..., int(np.ceil(com_slicewise)) :] = self._MEDIAL_KEY
        regions_volume += ml_volume

        # deep/superficial division
        rho_volume = np.stack([rho] * num_slices, axis=-1)
        deep_volume = (rho_volume <= rhos_threshold_volume) * self._DEEP_KEY
        superficial_volume = (rho_volume >= rhos_threshold_volume) * self._SUPERFICIAL_KEY
        ds_volume = np.asarray(
            deep_volume + superficial_volume + self._TOTAL_AXIAL_KEY, dtype=np.uint16
        )

        regions_volume += ds_volume
        ml_boundary = int(np.ceil(com_slicewise))
        acp_boundary = [
            int(np.floor((-105 - theta_min) / dtheta)),
            int(np.floor((-75 - theta_min) / dtheta)),
        ]

        return regions_volume, theta_bins, ml_boundary, acp_boundary

    def unroll(self, qv_map: np.ndarray, regions_mask: np.ndarray, theta_bins):
        """Unroll femoral cartilage 3D quantitative value (qv) maps to 2D for visualization.

        The function multiplies a 3D segmentation mask to a 3D qv map to produce a 3D femoral
        cartilage qv (fc_qv) map. It then fits a circle to the collapsed sagittal projection
        of the fc_qv map. Each slice is binned into bins of 5 degree sizes

        The unrolled map is then divided into deep and superficial cartilage.

        Args:
            qv_map (np.ndarray): 3D array (slices last) of sagittal knee describing
                quantitative parameter values regions_mask (np.ndarray): regions_mask
        Returns:
            tuple: (row, column) format
                1. 2D Total unrolled cartilage (slices, degrees) - average of superficial
                    and deep layers
                2. Superficial unrolled cartilage (slices, degrees) - superficial layer
                3. Deep unrolled cartilage (slices, degrees) - deep layer
        """
        num_bins = len(np.unique(theta_bins))

        mask = self.__mask__.volume

        if qv_map.shape != mask.shape:
            raise ValueError("t2_map and mask must have same shape")

        if len(qv_map.shape) != 3:
            raise ValueError("t2_map and mask must be 3D")

        # assert self.regions_mask is not None, (
        #     "region_mask not initialized. Should be initialized when mask is set"
        # )

        num_slices = qv_map.shape[-1]

        qv_map = np.nan_to_num(qv_map)
        qv_map = np.multiply(mask, qv_map)  # apply binary mask
        qv_map[
            qv_map <= 0
        ] = np.nan  # wherever qv_map is 0, either no cartilage or qv=0 ms, which is impractical

        # theta_bins = self.theta_bins  # binning with theta

        # regions_mask = self.regions_mask

        Unrolled_Cartilage = np.zeros([num_bins, num_slices])
        Sup_layer = np.zeros([num_bins, num_slices])
        Deep_layer = np.zeros([num_bins, num_slices])

        for slice_ind in range(num_slices):
            qv_slice = qv_map[..., slice_ind]
            curr_slice = regions_mask[..., slice_ind]

            # if slice is all NaNs, then don't analyze
            if np.sum(np.isnan(qv_slice)) == qv_slice.shape[0] * qv_slice.shape[1]:
                continue

            for curr_bin in range(num_bins):
                qv_bin = qv_slice[theta_bins == curr_bin]
                if np.sum(np.isnan(qv_bin)) == len(qv_bin):
                    continue

                Unrolled_Cartilage[curr_bin, slice_ind] = np.nanmean(qv_bin)

                qv_superficial = qv_slice[
                    np.logical_and(
                        theta_bins == curr_bin,
                        self.__binarize_region_mask__(curr_slice, self._SUPERFICIAL_KEY),
                    )
                ]
                qv_deep = qv_slice[
                    np.logical_and(
                        theta_bins == curr_bin,
                        self.__binarize_region_mask__(curr_slice, self._DEEP_KEY),
                    )
                ]

                qv_superficial = np.nan_to_num(qv_superficial)
                qv_deep = np.nan_to_num(qv_deep)

                qv_sup_mean = np.mean(qv_superficial[qv_superficial > 0])
                qv_deep_mean = np.mean(qv_deep[qv_deep > 0])
                Sup_layer[curr_bin, slice_ind] = qv_sup_mean
                Deep_layer[curr_bin, slice_ind] = qv_deep_mean

        Unrolled_Cartilage[Unrolled_Cartilage == 0] = np.nan
        Sup_layer[Sup_layer == 0] = np.nan
        Deep_layer[Deep_layer == 0] = np.nan

        return Unrolled_Cartilage, Sup_layer, Deep_layer

    def __calc_quant_vals__(self, quant_map: MedicalVolume, map_type):
        """Calculate quantitative values per region and 2D visualizations

        1. Save 2D figure (deep, superficial, total) information to use with matplotlib
            (title, data, xlabel, ylabel, filename)

        2. Save 2D dataframes in format
                [['DMA', 'DMC', 'DMP'], ['DLA', 'DLC', 'DLP'],
                 ['SMA', 'SMC', 'SMP'], ['SLA', 'SLC', 'SLP'],
                 ['TMA', 'TMC', 'TMP'], ['TLA', 'TLC', 'TLP']]

                 D=deep, S=superficial, T=total,
                 M=medial, L=lateral,
                 A=anterior, C=central, P=posterior

        Args:
            quant_map (MedicalVolume): 3D volumes of quantitative values.
                Volume should have ``np.nan`` values for all pixels unable to be calculated.
            map_type (QuantitativeValueType): Type of quantitative value to analyze.
        """

        super().__calc_quant_vals__(quant_map, map_type)

        # assert self.regions_mask is not None, (
        #     "region_mask not initialized. Should be initialized when mask is set"
        # )

        # We have to call this every time we load a new quantitative map
        # mask = segmentation_mask * clipped_quant_map
        regions_mask, theta_bins, ml_boundary, acp_boundary = self.split_regions(quant_map.volume)
        if self.ML_BOUNDARY is None:
            self.ML_BOUNDARY = ml_boundary
        if self.ACP_BOUNDARY is None:
            self.ACP_BOUNDARY = acp_boundary

        total, superficial, deep = self.unroll(quant_map.volume, regions_mask, theta_bins)

        assert total.shape == deep.shape
        assert deep.shape == superficial.shape

        # regions_mask = self.regions_mask
        mask = self.__mask__.volume

        subject_pid = self.pid
        pd_header = ["Subject", "Location", "Side", "Region", "Mean", "Std", "Median", "# Voxels"]
        pd_list = []

        # Replace strings with values - eg. DMA = 'deep, medial, anterior'
        # tissue_values = [['DMA', 'DMC', 'DMP'], ['DLA', 'DLC', 'DLP'],
        #                  ['SMA', 'SMC', 'SMP'], ['SLA', 'SLC', 'SLP'],
        #                  ['TMA', 'TMC', 'TMP'], ['TLA', 'TLC', 'TLP']]
        # tissue_values = []

        for axial_ind in range(len(self._AXIAL_KEYS)):
            axial = self._AXIAL_KEYS[axial_ind]

            for sagittal_ind in range(len(self._SAGITTAL_KEYS)):
                sagittal = self._SAGITTAL_KEYS[sagittal_ind]
                for coronal_ind in range(len(self._CORONAL_KEYS)):
                    coronal = self._CORONAL_KEYS[coronal_ind]

                    curr_region_mask = self.__binarize_region_mask__(
                        regions_mask, (axial | coronal | sagittal)
                    )
                    curr_region_mask = curr_region_mask * mask * quant_map.volume

                    # discard all values that are <= 0
                    qv_region_vals = curr_region_mask[curr_region_mask > 0]

                    num_voxels = len(qv_region_vals)

                    c_mean = np.nanmean(qv_region_vals)
                    c_std = np.nanstd(qv_region_vals)
                    c_median = np.nanmedian(qv_region_vals)

                    row_info = [
                        subject_pid,
                        self._AXIAL_NAMES[axial_ind],
                        self._SAGITTAL_NAMES[sagittal_ind],
                        self._CORONAL_NAMES[coronal_ind],
                        c_mean,
                        c_std,
                        c_median,
                        num_voxels,
                    ]

                    pd_list.append(row_info)

        df = pd.DataFrame(pd_list, columns=pd_header)
        qv_name = map_type.name
        maps = [
            {
                "title": "{} deep".format(qv_name),
                "data": deep,
                "xlabel": "Slice",
                "ylabel": "Angle (binned)",
                "filename": "{}_deep".format(qv_name),
                "raw_data_filename": "{}_deep.data".format(qv_name),
            },
            {
                "title": "{} superficial".format(qv_name),
                "data": superficial,
                "xlabel": "Slice",
                "ylabel": "Angle (binned)",
                "filename": "{}_superficial".format(qv_name),
                "raw_data_filename": "{}_superficial.data".format(qv_name),
            },
            {
                "title": "{} total".format(qv_name),
                "data": total,
                "xlabel": "Slice",
                "ylabel": "Angle (binned)",
                "filename": "{}_total".format(qv_name),
                "raw_data_filename": "{}_total.data".format(qv_name),
            },
        ]

        self.__store_quant_vals__(maps, df, map_type)

    def set_mask(
        self, mask: MedicalVolume, use_largest_cc: bool = True, split_regions: bool = True
    ):
        """Set mask for tissue.

        Mask is cleaned by selecting the largest connected component from the mask.
            Femoral cartilage is expected to be single connected tissue.

        Args:
            mask (MedicalVolume): Binary mask of segmented tissue.
        """
        xp = get_array_module(mask.A)
        if use_largest_cc:
            msk = xp.asarray(largest_cc(mask.A), dtype=xp.uint8)
        else:
            msk = xp.asarray(mask.A, dtype=xp.uint8)
        mask_copy = mask._partial_clone(volume=msk)

        super().set_mask(mask_copy)

        if split_regions:
            (
                self.regions_mask,
                self.theta_bins,
                self.ML_BOUNDARY,
                self.ACP_BOUNDARY,
            ) = self.split_regions(  # noqa: E501
                self.__mask__.volume
            )

    def __save_quant_data__(self, dirpath: str):
        """Save quantitative data and 2D visualizations of femoral cartilage.

        Check which quantitative values (T2, T1rho, etc) are defined for femoral cartilage
        and analyze these:

            1. Save 2D total, superficial, and deep visualization maps.
            2. Save {'medial', 'lateral'}, {'anterior', 'central', 'posterior'},
            q{'deep', 'superficial'} data to excel file

        Args:
            dirpath (str): Directory path to tissue data.
        """
        q_names = []
        dfs = []

        for quant_val in QuantitativeValueType:
            if quant_val.name not in self.quant_vals.keys():
                continue

            q_names.append(quant_val.name)
            q_val = self.quant_vals[quant_val.name]
            dfs.append(q_val[1])

            q_name_dirpath = io_utils.mkdirs(os.path.join(dirpath, quant_val.name.lower()))
            for q_map_data in q_val[0]:
                filepath = os.path.join(q_name_dirpath, q_map_data["filename"])
                xlabel = "Slice"
                ylabel = "Angle (binned)"
                title = q_map_data["title"]
                data_map = q_map_data["data"]

                plt.clf()

                upper_bound = BOUNDS[quant_val]

                if preferences.visualization_use_vmax:
                    # Hard bounds - clipping
                    plt.imshow(data_map, cmap="jet", vmin=0.0, vmax=BOUNDS[quant_val])
                else:
                    # Try to use a soft bounds
                    if np.sum(data_map <= upper_bound) == 0:
                        plt.imshow(data_map, cmap="jet", vmin=0.0, vmax=BOUNDS[quant_val])
                    else:
                        warnings.warn(
                            "%s: Pixel value exceeded upper bound (%0.1f). Using normalized scale."
                            % (quant_val.name, upper_bound)
                        )
                        plt.imshow(data_map, cmap="jet")

                plt.xlabel(xlabel)
                plt.ylabel(ylabel)
                plt.title(title)
                clb = plt.colorbar()
                clb.ax.set_title("(ms)")

                plt.savefig(filepath)

                # Save data
                raw_data_filepath = os.path.join(
                    q_name_dirpath, "raw_data", q_map_data["raw_data_filename"]
                )
                io_utils.save_pik(raw_data_filepath, data_map)

        if len(dfs) > 0:
            io_utils.save_tables(os.path.join(dirpath, "data.xlsx"), dfs, q_names)

    def save_data(self, save_dirpath, data_format: ImageDataFormat = preferences.image_data_format):
        super().save_data(save_dirpath, data_format=data_format)

        save_dirpath = self.__save_dirpath__(save_dirpath)

        if self.regions_mask is None:
            return

        sagital_region_mask, coronal_region_mask = self.__split_mask__()

        # Save region map - add by 1 because no key can be 0
        coronal_region_mask = (coronal_region_mask + 1) * 10
        sagital_region_mask = sagital_region_mask + 1
        joined_mask = coronal_region_mask + sagital_region_mask
        labels = [
            "medial posterior",
            "medial central",
            "medial anterior",
            "lateral posterior",
            "lateral central",
            "lateral anterior",
        ]
        plt_dict = {
            "labels": labels,
            "xlabel": "Slice",
            "ylabel": "Angle (binned)",
            "title": "Unrolled Regions",
        }
        img_utils.write_regions(
            os.path.join(save_dirpath, "region_map"), joined_mask, plt_dict=plt_dict
        )

    def __binarize_region_mask__(self, region_mask, roi):
        return np.asarray(np.bitwise_and(region_mask, roi) == roi, dtype=np.bool)

    def __split_mask__(self):
        assert (
            self.ML_BOUNDARY is not None and self.ACP_BOUNDARY is not None
        ), "medial/lateral and anterior/central/posterior boundaries should be specified"

        # split into regions
        unrolled_total, _, _ = self.unroll(
            np.asarray(self.__mask__.volume, dtype=np.float32), self.regions_mask, self.theta_bins
        )

        acp_division_unrolled = np.zeros(unrolled_total.shape)

        ac_threshold = self.ACP_BOUNDARY[0]
        cp_threshold = self.ACP_BOUNDARY[1]
        acp_division_unrolled[:ac_threshold, :] = self._ANTERIOR_KEY
        acp_division_unrolled[ac_threshold:cp_threshold, :] = self._CENTRAL_KEY
        acp_division_unrolled[cp_threshold:, :] = self._POSTERIOR_KEY

        ml_division_unrolled = np.zeros(unrolled_total.shape)
        if self.medial_to_lateral:
            ml_division_unrolled[..., : self.ML_BOUNDARY] = self._MEDIAL_KEY
            ml_division_unrolled[..., self.ML_BOUNDARY :] = self._LATERAL_KEY
        else:
            ml_division_unrolled[..., : self.ML_BOUNDARY] = self._LATERAL_KEY
            ml_division_unrolled[..., self.ML_BOUNDARY :] = self._MEDIAL_KEY

        acp_division_unrolled[np.isnan(unrolled_total)] = np.nan
        ml_division_unrolled[np.isnan(unrolled_total)] = np.nan

        return acp_division_unrolled, ml_division_unrolled



"""Default parameters.

This is the first file that is imported from this software.
All initialization imports should occur here.

Please change depending on your preferences.
"""
import os
import shutil
import warnings
import yaml
from typing import Any, Mapping

from dosma.utils import env

import matplotlib
import nested_lookup

__all__ = ["preferences"]

# Parse preferences file
_resources_dir = env.resources_dir()
_internal_preferences_template_filepath = os.path.join(_resources_dir, "templates/.preferences.yml")
_preferences_cmd_line_filepath = os.path.join(
    _resources_dir, "templates/.preferences_cmd_line_schema.yml"
)

__preferences_filepath__ = os.path.join(_resources_dir, "preferences.yml")

if not os.path.isfile(__preferences_filepath__):
    shutil.copyfile(_internal_preferences_template_filepath, __preferences_filepath__)

# Default rounding for I/O (dicom, nifti, etc) - DO NOT CHANGE
AFFINE_DECIMAL_PRECISION = 4
SCANNER_ORIGIN_DECIMAL_PRECISION = 4

DEFAULT_FONT_SIZE = 16
DEFAULT_TEXT_SPACING = DEFAULT_FONT_SIZE * 0.01


class _Preferences(object):
    """A pseudo-Singleton class implementation to track preferences.

    Do not instantiate this class. To modify/update preferences use the ``preferences``
    module variable defined below.

    However, in the case this class is instantiated, a new object will be created,
    but all config properties will be shared among all instances. Meaning modifying
    the config in one instance will impact the preferences state in another instance.
    """

    _template_filepath = _internal_preferences_template_filepath
    _preferences_filepath = __preferences_filepath__
    __config = {}
    __key_list = []

    def __init__(self):
        # Load config and set information if config has not been initialized.
        if not self.__config:
            with open(self._preferences_filepath, "r") as f:
                self.__config = yaml.safe_load(f)
            with open(self._template_filepath, "r") as f:
                template = yaml.safe_load(f)

            self.__config = self._merge_dicts(self.__config, template)

            matplotlib.rcParams.update(self.__config["visualization"]["matplotlib"]["rcParams"])

            # Store all preference keys.
            self.__key_list = self._unroll_keys(self.config, "")

    def _merge_dicts(self, target, base, prefix=""):
        """Merges dicts from target onto base."""
        target_keys = target.keys()
        base_keys = base.keys()
        all_keys = target_keys | base_keys

        for k in all_keys:
            prefix_k = f"{prefix}/{k}" if prefix else k

            if k in target_keys and k not in base_keys:
                # Allow target config to specify parameters not in the base config.
                # Note this may change in the future.
                pass
            elif k not in target_keys and k in base_keys:
                warnings.warn(
                    "Your preferences file may be outdated. "
                    "Run `preferences.save()` to save your updated preferences."
                )
                target[k] = base[k]
            else:
                target_v = target[k]
                template_v = base[k]

                if isinstance(target_v, Mapping) and isinstance(template_v, Mapping):
                    target[k] = self._merge_dicts(target_v, template_v, prefix=prefix_k)
                elif isinstance(template_v, Mapping):
                    raise ValueError(
                        f"{prefix_k}: Got target type '{type(target_v)}', "
                        f"but base config type is '{type(template_v)}'"
                    )
                else:
                    # If both are not mapping, keep the target value.
                    pass

        return target

    def _unroll_keys(self, subdict: dict, prefix: str) -> list:
        """Recursive method to unroll keys."""
        pref_keys = []
        keys = subdict.keys()
        for k in keys:
            prefix_or_key = "{}/{}".format(prefix, k)
            val = subdict[k]
            if type(val) is dict:
                pref_keys.extend(self._unroll_keys(val, prefix_or_key))
            else:
                pref_keys.append(prefix_or_key)

        return pref_keys

    @staticmethod
    def _get_prefixes(prefix: str = ""):
        if not prefix:
            return ()

        r_prefixes = []
        prefixes = prefix.split("/")
        for p in prefixes:
            if p == "":
                continue
            r_prefixes.append(p)

        return r_prefixes

    def get(self, key: str, prefix: str = ""):
        """Get preference.

        Args:
            key (str): Preference to peek. Can be full path preference.
            prefix (str, optional): Prefix defining which sub-config to search.
                For example, to access the key ``"visualization/rcParams"``, this parameter
                should be ``"visualization"``.

        Returns:
            The preference value.
        """
        return self.__get(self.__config, key, prefix)[0]

    def __get(self, b_dict: dict, key: str, prefix: str = ""):
        """Get preference.

        Args:
            b_dict (dict): Dictionary to search.
            key (str): Preference to peek. Can be full path preference.
            prefix (:obj:`str`, optional): Prefix defining which sub-config to search.
                For example, to access the key ``"visualization/rcParams"``,
                this parameter should be ``"visualization"``.

        Returns:
            The preference value, sub-dictionary to search
        """
        p_keys = self._get_prefixes(key)
        key = p_keys[-1]
        k_prefixes = list(p_keys[:-1])

        prefixes = list(self._get_prefixes(prefix))
        prefixes.extend(k_prefixes)

        subdict = b_dict
        for p in prefixes:
            subdict = subdict[p]

        num_keys = nested_lookup.get_occurrence_of_key(subdict, key)

        if num_keys == 0:
            raise KeyError("Key not '%s ' found in prefix '%s'" % (key, prefix))
        if num_keys > 1:
            raise KeyError(
                "Multiple keys '%s 'found in prefix '%s'. Provide a more specific prefix."
                % (key, prefix)
            )

        return nested_lookup.nested_lookup(key, subdict)[0], subdict

    def set(self, key: str, value: Any, prefix: str = ""):
        """Set preference.

        Args:
            key (str): Preference to peek. Can be full path preference.
            value (Any): Value to set preference.
            prefix (str, optional): Prefix defining which sub-config to search.
                For example, to access the key ``visualization/rcParams``,
                this parameter should be `visualization`. Defaults to ''.

        Returns:
            The preference value.
        """
        val, subdict = self.__get(self.__config, key, prefix)

        # type of new value has to be the same type as old value
        if type(value) != type(val):
            try:
                value = type(val)(value)
            except (ValueError, TypeError):
                raise TypeError("Could not convert %s to %s: %s" % (type(value), type(val), value))

        p_keys = self._get_prefixes(key)
        key = p_keys[-1]
        nested_lookup.nested_update(subdict, key, value, in_place=True)

        # if param is an rcParam, update matplotlib
        if key in self.config["visualization"]["matplotlib"]["rcParams"].keys():
            matplotlib.rcParams.update({key: value})

    def save(self):
        """Save preferences to file.

        Args:
            file_path (str, optional): File path to yml file.
                Defaults to local preferences file.
        """
        with open(self._preferences_filepath, "w") as f:
            yaml.dump(self.__config, f, default_flow_style=False)

    @property
    def config(self):
        """Get preferences configuration."""
        return self.__config

    # Make certain preferences easily accessible through this class.
    @property
    def segmentation_batch_size(self) -> int:
        """int: Batch size for segmentation models."""
        return self.get("/segmentation/batch.size")

    @property
    def visualization_use_vmax(self) -> bool:
        return self.get("/visualization/use.vmax")

    @property
    def mask_dilation_rate(self) -> float:
        """float: rate for mask dilation."""
        return self.get("/registration/mask/dilation.rate")

    @property
    def mask_dilation_threshold(self) -> float:
        """float: threshold for mask dilation."""
        return self.get("/registration/mask/dilation.threshold")

    @property
    def fitting_r2_threshold(self) -> float:
        """float: r2 threshold for fitting"""
        return self.get("/fitting/r2.threshold")

    @property
    def image_data_format(self):
        """ImageDataFormat: Format for images (e.g. png, eps, etc.)."""
        from dosma.core.io.format_io import ImageDataFormat

        return ImageDataFormat[self.get("/data/format")]

    @property
    def nipype_logging(self) -> str:
        """str: nipype library logging mode."""
        # TODO: Remove this try/except clause when schema is well defined.
        try:
            return self.get("/logging/nipype")
        except KeyError:
            return "file_stderr"

    def cmd_line_flags(self) -> dict:
        """Provide command line flags for changing preferences via command line.

        Not all preferences are listed here. Only those that should easily be set.

        All default values will be based on the current state of preferences,
        not the static state specified in yaml file.

        Returns:
            Preference keys with corresponding argparse kwarg dict
        """
        with open(_preferences_cmd_line_filepath) as f:
            cmd_line_config = yaml.safe_load(f)

        cmd_line_dict = {}
        for k in self.__key_list:
            try:
                argparse_kwargs, _ = self.__get(cmd_line_config, k)
            except KeyError:
                continue

            argparse_kwargs["default"] = self.get(k)
            argparse_kwargs["type"] = eval(argparse_kwargs["type"])
            cmd_line_dict[k] = argparse_kwargs

        return cmd_line_dict

    def __str__(self):
        return str(self.__config)


preferences = _Preferences()



import os
from abc import ABC, abstractmethod
from typing import Union

import numpy as np
import pandas as pd
import scipy.ndimage as sni

from dosma.core.io import format_io_utils as fio_utils
from dosma.core.io.format_io import ImageDataFormat
from dosma.core.med_volume import MedicalVolume
from dosma.core.orientation import SAGITTAL
from dosma.core.quant_vals import QuantitativeValue, QuantitativeValueType
from dosma.defaults import preferences
from dosma.utils import io_utils

WEIGHTS_FILE_EXT = "h5"

__all__ = ["Tissue"]


class Tissue(ABC):
    """Abstract class for tissues.

    Tissues are defined loosely as any tissue structures (bones, soft tissue, etc.).

    Args:
        weights_dir (str): Directory to all segmentation weights.
        medial_to_lateral (`bool`, optional): If `True`, anatomy is from medial_to_lateral.

    Attributes:
        FULL_NAME (str): Full name of tissue 'femoral cartilage' for femoral cartilage.
        ID (int): Unique integer ID for tissue. Should be unique to all tissues,
            and should not change.
        STR_ID (str): Short hand string id such as 'fc' for femoral cartilage.
        T1_EXPECTED (float): Expected T1 value (in milliseconds).
        medial_to_lateral (bool): If ``True``, mask is in medial to lateral direction.
        pid (str): Patient/subject ID. Should be anonymized.
        quant_vals (dict[str, tuple[np.ndarray, pd.DataFrame]]): Mapping from quantitative value
            name (t2, t1-rho, etc.) to tuple of unrolled map and DataFrame containing
            measurement values.
        weights_filepath (str): File path to weights directory for neural network segmentation.
    """

    ID = -1
    STR_ID = ""
    FULL_NAME = ""

    # Expected quantitative param values.
    T1_EXPECTED = None

    def __init__(self, weights_dir: str = None, medial_to_lateral: bool = None):
        self.pid = None
        self.__mask__ = None
        self.quant_vals = {}
        self.weights_file_path = None

        if weights_dir is not None:
            self.weights_file_path = self.find_weights(weights_dir)

        self.medial_to_lateral = medial_to_lateral

        # quantitative value list
        self.quantitative_values = []

    @abstractmethod
    def split_regions(self, base_map: Union[np.ndarray, MedicalVolume]):
        """Split mask into anatomical regions.

        Args:
            base_map (np.ndarray): 3D numpy array typically corresponding to volume to split.

        Returns:
            np.ndarray: 4D numpy array (region, height, width, depth).
                        Saved in variable `self.regions`.
        """
        pass

    def calc_quant_vals(self):
        """Calculate quantitative values for pixels corresponding to the tissue.

        Requires mask to be set for this tissue.
        """
        for qv in self.quantitative_values:
            self.__calc_quant_vals__(qv.volumetric_map, qv.qv_type)

    @abstractmethod
    def __calc_quant_vals__(self, quant_map: MedicalVolume, map_type: QuantitativeValueType):
        """Helper method to get quantitative values for tissue - implemented per tissue.

        Different tissues should override this as they see fit.

        Args:
            quant_map (MedicalVolume): 3D map of pixel-wise quantitative measures
                (T2, T2*, T1-rho, etc.). Volume should have ``np.nan`` values for
                all pixels unable to be calculated.
            map_type (QuantitativeValueType): Type of quantitative value to analyze.

        Raises:
            TypeError: If `quant_map` is not of type `MedicalVolume` or `map_type` is not of type
                `QuantitativeValueType`.
            ValueError: If no mask is found for tissue.
        """
        if not isinstance(quant_map, MedicalVolume):
            raise TypeError("`Expected type 'MedicalVolume' for `quant_map`")
        if not isinstance(map_type, QuantitativeValueType):
            raise TypeError("`Expected type 'QuantitativeValueType' for `map_type`")

        if self.__mask__ is None:
            raise ValueError("Please initialize mask for {}".format(self.FULL_NAME))

        quant_map.reformat(self.__mask__.orientation, inplace=True)
        pass

    def __store_quant_vals__(
        self, quant_map: MedicalVolume, quant_df: pd.DataFrame, map_type: QuantitativeValueType
    ):
        """Adds quantitative value in `self.quant_vals`.

        Args:
            quant_map (list[dict]): Dictionaries of different unrolled maps and
                corresponding plotting data (title, xlabel, etc.).
            quant_df (pd.DataFrame): Computed data for this quantitative value.
            map_type (QuantitativeValueType): Type of quantitative value to analyze.
        """
        self.quant_vals[map_type.name] = (quant_map, quant_df)

    def find_weights(self, weights_dir: str):
        """Search for weights file in weights directory.

        Args:
            weights_dir (str): Directory where weights are stored.

        Returns:
            str: File path to weights corresponding to tissue.

        Raises:
            ValueError: If multiple weights files exists for the tissue
                or no valid weights file found.
        """

        # Find weights file with NAME in the filename, like 'fc_weights.h5'
        files = os.listdir(weights_dir)
        weights_file = None
        for f in files:
            file = os.path.join(weights_dir, f)
            if os.path.isfile(file) and f.endswith(WEIGHTS_FILE_EXT) and self.STR_ID in f:
                if weights_file is not None:
                    raise ValueError("There are multiple weights files, please remove duplicates")
                weights_file = file

        if weights_file is None:
            raise ValueError(
                "No file found that contains '{}' and ends in '{}'".format(
                    self.STR_ID, WEIGHTS_FILE_EXT
                )
            )

        self.weights_file_path = weights_file

        return weights_file

    def save_data(
        self, save_dirpath: str, data_format: ImageDataFormat = preferences.image_data_format
    ):
        """Save data for tissue.

        Saves mask and quantitative values associated with this tissue.

        Override in subclasses to save additional data. When overriding in subclasses, call
        ``super().save_data(save_dirpath)`` first to save mask and quantitative values by default.
        See :mod:`dosma.tissues.femoral_cartilage` for details.

        .. literalinclude:: femoral_cartilage.py

        Args:
            save_dirpath (str): Directory path where all data is stored.
            data_format (`ImageDataFormat`, optional): Format to save data.
        """
        save_dirpath = self.__save_dirpath__(save_dirpath)

        if self.__mask__ is not None:
            mask_file_path = os.path.join(save_dirpath, "{}.nii.gz".format(self.STR_ID))
            mask_file_path = fio_utils.convert_image_data_format(mask_file_path, data_format)
            self.__mask__.save_volume(mask_file_path, data_format=data_format)

        for qv in self.quantitative_values:
            qv.save_data(save_dirpath, data_format)

        self.__save_quant_data__(save_dirpath)

    @abstractmethod
    def __save_quant_data__(self, dirpath: str):
        """Save quantitative data generated for this tissue.

        Called by `save_data`.

        Args:
            dirpath (str): Directory path to tissue data.
        """
        pass

    def save_quant_data(self, dirpath: str):
        """Save quantitative data generated for this tissue.

        Does not save mask or quantitative parameter map.

        Args:
            dirpath (str): Directory path to tissue data.
        """
        return self.__save_quant_data__(dirpath)

    def load_data(self, load_dir_path: str):
        """Load data for tissue.

        All tissue information is based on the mask. If mask for tissue doesn't exist,
        there is no information to load.

        Args:
            load_dir_path (str): Directory path where all data is stored.
        """
        load_dir_path = self.__save_dirpath__(load_dir_path)
        mask_file_path = os.path.join(load_dir_path, "{}.nii.gz".format(self.STR_ID))

        # Try to load mask, if file exists.
        try:
            msk = fio_utils.generic_load(mask_file_path, expected_num_volumes=1)
            self.set_mask(msk)
        except FileNotFoundError:
            # do nothing
            pass

        self.quantitative_values = QuantitativeValue.load_qvs(load_dir_path)

    def __save_dirpath__(self, dirpath):
        """Tissue-specific subdirectory to store data.

        Subdirectory will have path '`dirpath`/`self.STR_ID`/'.

        If directory does not exist, it will be created.

        Args:
            dirpath (str): Directory path where all data is stored.

        Returns:
            str: Tissue-specific data directory.
        """
        return io_utils.mkdirs(os.path.join(dirpath, self.STR_ID))

    # TODO (arjundd): Refactor get/set methods of mask to property.
    def set_mask(self, mask: MedicalVolume):
        """Set mask for tissue.

        Args:
            mask (MedicalVolume): Binary mask of segmented tissue.
        """
        assert type(mask) is MedicalVolume, "mask for tissue must be of type MedicalVolume"
        mask = mask.reformat(SAGITTAL)
        self.__mask__ = mask

    def get_mask(self):
        """
        Returns:
            MedicalVolume: Binary mask of segmented tissue.
        """
        return self.__mask__

    def add_quantitative_value(self, qv_new: QuantitativeValue):
        """Add quantitative value to the tissue.

        Args:
            qv_new (QuantitativeValue): Quantitative value to add to tissue.
        """
        # for qv in self.quantitative_values:
        #     if qv_new.NAME == qv.NAME:
        #         raise ValueError('This quantitative value already exists. '
        #                          'Only one type of quantitative value can be added per tissue.\n'
        #                          'Manually delete %s folder' % qv_new.NAME)

        self.quantitative_values.append(qv_new)

    def __get_axis_bounds__(
        self, im: np.ndarray, ignore_nan: bool = True, leave_buffer: bool = False
    ):
        """Get tightest bounds for data in the array.

        When plotting data, we would like to avoid making our dynamic range too large such
        that we cannot detect color changes in differences that matter.
        To avoid this, we make our bounds as tight as possible.

        Bounds are calculated with respect to non-zero elements. If unique values are [0, 8, 9],
            the dyanmic range will be [8, 9].

        Args:
            im (np.ndarray): Array containing information for which bounds have to be computed.
            ignore_nan (obj:`bool`, optional): Ignore `nan` values when computing the bounds.
            leave_buffer (obj:`bool`, optional): Add buffer of +/-5 to dynamic range.
        """
        im_temp = im
        axs = []
        if ignore_nan:
            im_temp = np.nan_to_num(im)

        non_zero_elems = np.nonzero(im_temp)

        for i in range(len(non_zero_elems)):
            v_min = np.min(non_zero_elems[i])
            v_max = np.max(non_zero_elems[i])
            if leave_buffer:
                v_min -= 5
                v_max += 5

            axs.append((v_min, v_max))

        return axs


def largest_cc(mask, num=1):
    """Return the largest `num` connected component(s) of a 3D mask array.

    Args:
        mask (np.ndarray): 3D mask array (`np.bool` or `np.[u]int`).
        num (int, optional): Maximum number of connected components to keep.

    Returns:
        mask (np.ndarray): 3D mask array with `num` connected components.


    Note:
        Adapted from nipy (https://github.com/nipy/nipy/blob/master/nipy/labs/mask.py)
        due to dependency issues.
    """
    # We use asarray to be able to work with masked arrays.
    mask = np.asarray(mask)
    labels, label_nb = sni.label(mask)
    if not label_nb:
        raise ValueError("No non-zero values: no connected components")
    if label_nb == 1:
        return mask.astype(np.bool)
    label_count = np.bincount(labels.ravel().astype(np.int))
    # discard 0 the 0 label
    label_count[0] = 0

    # Split num=1 case for speed.
    if num == 1:
        return labels == label_count.argmax()
    else:
        # 1) discard 0 the 0 label and 2) descending order
        order = np.argsort(label_count)[1:][::-1]
        return np.isin(labels, order[:num])



import itertools
import logging
import multiprocessing as mp
import os
import platform
import shutil
import subprocess
import sys
import uuid
import warnings
from functools import partial
from typing import Dict, Sequence, Union

from nipype.interfaces.elastix import ApplyWarp, Registration
from nipype.interfaces.elastix.registration import RegistrationOutputSpec
from tqdm import tqdm
from tqdm.contrib.concurrent import process_map

from dosma import file_constants as fc
from dosma.core.device import cpu_device
from dosma.core.io.nifti_io import NiftiReader, NiftiWriter
from dosma.core.med_volume import MedicalVolume
from dosma.defaults import preferences
from dosma.utils import env

__all__ = ["register", "apply_warp", "symlink_elastix", "unlink_elastix"]

MedVolOrPath = Union[MedicalVolume, str]
_logger = logging.getLogger(__name__)


def register(
    target: MedVolOrPath,
    moving: Union[MedVolOrPath, Sequence[MedVolOrPath]],
    parameters: Union[str, Sequence[str]],
    output_path: str,
    target_mask: MedVolOrPath = None,
    moving_masks: Union[MedVolOrPath, Sequence[MedVolOrPath]] = None,
    sequential: bool = False,
    collate: bool = True,
    num_workers: int = 0,
    num_threads: int = 1,
    show_pbar: bool = False,
    return_volumes: bool = False,
    rtype: type = dict,
    **kwargs,
):
    """Register moving image(s) to the target.

    ``MedVolOrPath`` is a shorthand for ``Union[MedicalVolume, str, Path]``.
    It indicates the argument can be either a ``MedicalVolume`` or a path to a nifti file.

    Args:
        target (`MedicalVolume` or `str`): The target/fixed image.
        moving (MedicalVolume(s) or str(s)): The moving/source image(s).
        parameters (str(s)): Elastix parameter files to use.
        output_path (str): Output directory to store files.
        target_mask (MedicalVolume or str, optional): The target/fixed mask.
        moving_masks (MedicalVolume(s) or str(s), optional): The moving mask(s).
            If only one specified, the mask will be used for all moving images.
        sequential (bool, optional): If `True`, apply parameter files sequentially.
        collate (bool, optional): If `True`, will collate outputs from sequential registration
            into single RegistrationOutputSpec instance. If `sequential=False`, this argument
            is ignored.
        num_workers (int, optional): Number of workers to use for reading and writing data and
            for registration.
        num_threads (int, optional): Number of threads to use for registration.
            Note total number of threads used will be ``num_workers * num_threads``.
        show_pbar (bool, optional): If `True`, show progress bar during registration.
            Note the progress bar will not be shown for intermediate reading/writing.
        return_volumes (bool, optional): If `True`, registered volumes will also be returned.
            By default, only the output namespaces (RegistrationOutputSpec) of the registrations are
            returned.
        rtype (type, optional): The return type. Either `dict` or `tuple`.
        kwargs: Keyword arguments used to initialize `nipype.interfaces.elastix.Registration`.

    Returns:
        Dict or Tuple:
            Type specified by ``rtype``. If ``rtype=dict``, returns dict with keys ``'outputs'``
            and ``'volumes'`` (if ``return_volumes=True``). If ``rtype=tuple``, returns
            ``(outputs, volumes or None)``. Length of ``outputs`` and ``volumes`` depends on
            number of images specified in ``moving``:

            outputs (Sequence[RegistrationOutputSpec]): The output objects from
            elastix registration, one for each moving image. Each object is effectively
            a namespace with four main attributes:

                - 'transform' (List[str]): Paths to transform files produced using registration.
                - 'warped_file' (str): Path to the final registered image.
                - 'warped_files' (List[str]): Paths to all intermediate images created
                  if multiple parameter files used.

            volumes (Sequence[MedicalVolume]): Registered volumes.
    """
    assert issubclass(rtype, (Dict, Sequence))  # `rtype` must be dict or tuple
    has_output_path = bool(output_path)
    if not output_path:
        output_path = os.path.join(
            env.temp_dir(), f"register-{str(uuid.uuid1())}-{str(uuid.uuid4())}"
        )

    moving = [moving] if isinstance(moving, (MedicalVolume, str)) else moving
    moving_masks = (
        [moving_masks]
        if moving_masks is None or isinstance(moving_masks, (MedicalVolume, str))
        else moving_masks
    )
    if len(moving_masks) > 1 and len(moving) != len(moving_masks):
        raise ValueError(
            "Got {} moving images but {} moving masks".format(len(moving), len(moving_masks))
        )

    files = [target, target_mask] + moving + moving_masks
    if any(isinstance(f, MedicalVolume) and f.device != cpu_device for f in files):
        raise RuntimeError("MedicalVolume data must be on CPU")

    # Write medical volumes (if any) to nifti file for use with elastix.
    tmp_dir = os.path.join(output_path, "tmp")
    default_files = (
        ["target", "target-mask"]
        + [f"moving-{idx}" for idx in range(len(moving))]
        + [f"moving-mask-{idx}" for idx in range(len(moving_masks))]
    )  # noqa
    assert len(default_files) == len(files), default_files  # should be 1-to-1 with # args provided
    vols = [(idx, v) for idx, v in enumerate(files) if isinstance(v, MedicalVolume)]
    idxs, vols = [x[0] for x in vols], [x[1] for x in vols]

    # Temporary directory must be created prior to writing data
    # due to issues with creating directories in multiprocessing settings.
    os.makedirs(tmp_dir, exist_ok=True)

    if len(vols) > 0:
        filepaths = [os.path.join(tmp_dir, f"{default_files[idx]}.nii.gz") for idx in idxs]
        if num_workers > 0:
            with mp.Pool(min(num_workers, len(vols))) as p:
                out = p.starmap_async(_write, zip(vols, filepaths))
                out.wait()
        else:
            for vol, fp in zip(vols, filepaths):
                _write(vol, fp)
        for idx, fp in zip(idxs, filepaths):
            files[idx] = fp

    # Assign file paths to respective variables.
    target, moving = files[0], files[2 : 2 + len(moving)]
    target_mask, moving_masks = files[1], files[2 + len(moving) :]
    if len(moving_masks) == 1:
        moving_masks = moving_masks * len(moving)

    all_outputs = {}

    # Perform registration.
    reg_out_paths = [os.path.join(output_path, f"moving-{idx}") for idx in range(len(moving))]
    reg_args = list(zip(moving, moving_masks, reg_out_paths))
    if num_workers > 0:
        func = partial(
            _elastix_register_mp,
            target=target,
            parameters=parameters,
            target_mask=target_mask,
            sequential=sequential,
            collate=collate,
            num_threads=num_threads,
            **kwargs,
        )
        max_workers = min(num_workers, len(reg_args))
        out = process_map(
            func, reg_args, max_workers=max_workers, tqdm_class=tqdm, disable=not show_pbar
        )
    else:
        out = []
        for mvg, mvg_mask, out_path in tqdm(reg_args, disable=not show_pbar):
            _out = _elastix_register(
                target,
                mvg,
                parameters,
                out_path,
                target_mask,
                mvg_mask,
                sequential,
                collate,
                num_threads,
                **kwargs,
            )
            out.append(_out)

    all_outputs["outputs"] = tuple(out)

    # Load volumes.
    if return_volumes:
        filepaths = [x[-1].warped_file if isinstance(x, Sequence) else x.warped_file for x in out]
        if num_workers > 0:
            with mp.Pool(min(num_workers, len(filepaths))) as p:
                vols = p.map(_read, filepaths)
        else:
            vols = []
            for fp in filepaths:
                vols.append(_read(fp))
        all_outputs["volume"] = tuple(vols)

    # Clean up.
    for _dir in [tmp_dir, output_path if not has_output_path else None]:
        if not _dir or not os.path.isdir(_dir):
            continue
        shutil.rmtree(_dir)

    if issubclass(rtype, dict):
        out = rtype(all_outputs)
    elif issubclass(rtype, Sequence):
        out = rtype([all_outputs["outputs"], all_outputs.get("volume", None)])
    else:
        assert False  # Should have type checking earlier.

    return out


def apply_warp(
    moving: Union[MedVolOrPath, Sequence[MedVolOrPath]],
    transform: Union[str, Sequence[str]] = None,
    out_registration: RegistrationOutputSpec = None,
    output_path: Union[str, Sequence[str]] = None,
    rtype: type = MedicalVolume,
    num_threads: int = 1,
    show_pbar: bool = False,
    num_workers: int = 0,
) -> MedVolOrPath:
    """Apply transform(s) to moving image using transformix.

    Use transformix to apply a transform on an input image. The transform(s) is/are
    specified in the transform-parameter file(s).

    Args:
        moving (MedicalVolume(s) or str(s)): The moving/source image to transform.
        transform (str(s)): Paths to transform files to be used by transformix.
            If multiple files provided, transforms will be applied sequentially.
            If `None`, will be determined by `out_registration.transform`.
        out_registration (RegistrationOutputSpec(s)): Outputs from elastix registration
            using nipype. Must be specified if `transform` is None.
        output_path (str): Output directory to store files.
        rtype (type, optional): Return type - either `MedicalVolume` or `str`.
            If `str`, `output_path` must be specified. Defaults to `MedicalVolume`.
        num_threads (int, optional): Number of threads to use for registration.
            If `None`, defaults to 1.
        show_pbar (bool, optional): If `True`, show progress bar when applying transforms.

    Return:
        MedVolOrPath: The medical volume or nifti file corresponding to the volume.
            See `rtype` for details.
    """
    single_vol = isinstance(moving, (MedicalVolume, os.PathLike))
    if single_vol:
        if num_workers > 0:
            _logger.warning("Ignoring `num_workers` - only single volume was detected")
        return _apply_warp(
            moving=moving,
            transform=transform,
            out_registration=out_registration,
            output_path=output_path,
            rtype=rtype,
            num_threads=num_threads,
            show_pbar=show_pbar,
        )

    num_volumes = len(moving)
    seq_type = type(moving)

    if not output_path:
        output_path = [None] * num_volumes
    elif isinstance(output_path, (str, os.PathLike)):
        output_path = [os.path.join(output_path, f"image-{idx}") for idx in range(num_volumes)]
    elif not isinstance(output_path, Sequence) or len(output_path) != num_volumes:
        raise ValueError(
            "`output_path` must be a directory or list of directories " "of same length as `moving`"
        )

    warp_args = list(zip(moving, output_path))
    if num_workers > 0:
        func = partial(
            _apply_warp_mp,
            transform=transform,
            out_registration=out_registration,
            rtype=rtype,
            num_threads=num_threads,
            show_pbar=False,
        )
        max_workers = min(num_workers, len(warp_args))
        out = process_map(
            func, warp_args, max_workers=max_workers, tqdm_class=tqdm, disable=not show_pbar
        )
    else:
        out = []
        for mvg, out_path in tqdm(warp_args, disable=not show_pbar):
            _out = _apply_warp(
                moving=mvg,
                output_path=out_path,
                transform=transform,
                out_registration=out_registration,
                rtype=rtype,
                num_threads=num_threads,
                show_pbar=False,
            )
            out.append(_out)

    return seq_type(out)


def symlink_elastix(path: str = None, lib_only: bool = True, force: bool = False):
    """Symlinks elastix/transformix files to the dosma library.

    Args:
        path (str, optional): Path to elastix folder. This folder should
            contain two folders `bin` and `lib`. If `None`, determined
            using `which elastix`. This will overwrite existing linked
            files. path cannot be automatically determined on Windows.
        lib_only (bool, optional): If `True`, only links contents of `lib`
            folder.
        force (bool, optional): If `True`, unlinks existing files before relinking.
            Note this operation is not atomic.

    Note:
        Setting elastix paths this way is not recommended unless you
        are using a MacOS (Darwin) platform, where there are known
        path issues with elastix (https://github.com/almarklein/pyelastix/issues/9).
        For linux and windows machines, using the setup described in the elastix
        guide is sufficient.
    """
    system = platform.system().lower()
    assert system in ["windows", "darwin", "linux"]
    if system != "darwin":
        warnings.warn(
            f"Symlinking elastix/transformix paths not recommended for {system} " f"machines"
        )

    if path is None:
        if system == "windows":
            raise ValueError("`path` cannot be determined automatically on Windows")
        try:
            out = subprocess.check_output(["which", "elastix"]).decode("ascii").strip("\n")
            path = os.path.dirname(os.path.dirname(out))
        except subprocess.CalledProcessError:
            raise ValueError(
                "Path to `elastix` not intialized. "
                "Use `export PATH=/path/to/elastix/folder:$PATH`"
            )
    assert os.path.isdir(path), path  # must be a directory

    dirs = {"lib": [x for x in os.listdir(os.path.join(path, "lib")) if x.startswith("libANNlib")]}
    if not lib_only:
        dirs["bin"] = ["elastix", "transformix"]

    for dirname, files in dirs.items():
        for file in files:
            src = os.path.join(path, dirname, file)
            tgt = os.path.join(fc._DOSMA_ELASTIX_FOLDER, file)
            if os.path.exists(tgt):
                if force:
                    os.remove(tgt)
                else:
                    raise FileExistsError(
                        f"File {tgt} exists. "
                        f"Use `unlink_elastix` or `force` to unlink the file."
                    )
            os.symlink(src, tgt)


def unlink_elastix():
    """Unlinks all elastix/transformix files in the dosma library."""
    for x in os.listdir(fc._DOSMA_ELASTIX_FOLDER):
        x = os.path.join(fc._DOSMA_ELASTIX_FOLDER, x)
        if os.path.islink(x):
            os.remove(x)


def _elastix_register(
    target: str,
    moving: str,
    parameters: Sequence[str],
    output_path: str,
    target_mask: str = None,
    moving_mask: str = None,
    sequential=False,
    collate=True,
    num_threads=None,
    use_mask: Sequence[bool] = None,
    **kwargs,
):
    def _register(_moving, _parameters, _output_path, _use_mask=None):
        if isinstance(_parameters, str):
            _parameters = [_parameters]
        if _use_mask is None:
            _use_mask = target_mask is not None or moving_mask is not None

        _output_path = os.path.abspath(_output_path)
        os.makedirs(_output_path, exist_ok=True)

        elastix_path = _local_exe("elastix")
        cwd = _local_lib_dir()

        reg = Registration()
        if elastix_path:
            reg._cmd = elastix_path
        reg.inputs.fixed_image = os.path.abspath(target)
        reg.inputs.moving_image = os.path.abspath(_moving)
        reg.inputs.parameters = [os.path.abspath(p) for p in _parameters]
        reg.inputs.output_path = os.path.abspath(_output_path)
        reg.terminal_output = preferences.nipype_logging
        if num_threads:
            reg.inputs.num_threads = num_threads
        if _use_mask and target_mask is not None:
            reg.inputs.fixed_mask = os.path.abspath(target_mask)
        if _use_mask and moving_mask is not None:
            reg.inputs.moving_mask = os.path.abspath(moving_mask)
        for k, v in kwargs.items():
            setattr(reg.inputs, k, v)

        return reg.run(cwd=cwd).outputs

    def _collate_outputs(_outs):
        """
        Concatenates fields that are sequential and takes final output
        for fields that are not.
        """
        if len(_outs) == 1:
            return _outs[0]

        _result = _outs[0]
        fields = list(_outs[0].__dict__.keys())
        for _fld in fields:
            _res_val = getattr(_result, _fld)
            if not isinstance(_res_val, str) and isinstance(_res_val, Sequence):
                val = list(itertools.chain.from_iterable([getattr(x, _fld) for x in _outs]))
            else:
                val = getattr(_outs[-1], _fld)
            setattr(_result, _fld, val)
        return _result

    if use_mask is not None:
        assert sequential  # use_mask can only be specified when sequential is specified
    if sequential:
        outs, mvg = [], moving
        for idx, param in enumerate(parameters):
            _use_mask = None if use_mask is None else use_mask[idx]
            _out = _register(mvg, param, os.path.join(output_path, f"param{idx}"), _use_mask)
            outs.append(_out)
            mvg = _out.warped_file
        out = _collate_outputs(outs) if collate else outs
        return out
    else:
        return _register(moving, parameters, output_path)


def _elastix_register_mp(args, **kwargs):
    """Reorder arguments for multiprocessing support."""
    moving, moving_mask, output_path = args
    return _elastix_register(
        moving=moving, moving_mask=moving_mask, output_path=output_path, **kwargs
    )


def _apply_warp(
    moving: MedVolOrPath,
    output_path: str = None,
    transform: Union[str, Sequence[str]] = None,
    out_registration: RegistrationOutputSpec = None,
    rtype: type = MedicalVolume,
    num_threads: int = 1,
    show_pbar: bool = False,
) -> MedVolOrPath:
    assert rtype in [MedicalVolume, str], rtype  # rtype must be MedicalVolume or str
    has_output_path = bool(output_path)
    if rtype == str and not has_output_path:
        raise ValueError("`output_path` must be specified when `rtype=str`")
    if not output_path:
        # TODO: Add path generation that prevents collisions during multiprocessing.
        # When multiprocessing executes rapidly and poor seed is set, the uuids have
        # collided. To avoid this, we append the process name to the directory that is
        # created.
        output_path = os.path.join(
            env.temp_dir(), f"apply_warp-{str(uuid.uuid1())}-{str(uuid.uuid4())}"
        )
        if not _is_main_process():
            output_path += mp.current_process().name
    output_path = os.path.abspath(output_path)
    os.makedirs(output_path, exist_ok=True)

    if not transform:
        transform = out_registration.transform
    elif isinstance(transform, str):
        transform = [transform]
    transform = [os.path.abspath(t) for t in transform]

    mv_filepath = os.path.join(output_path, "moving.nii.gz")
    if isinstance(moving, MedicalVolume):
        NiftiWriter().save(moving, mv_filepath)
        moving = mv_filepath

    transformix_path = _local_exe("transformix")  # noqa
    cwd = _local_lib_dir()
    for tf in tqdm(transform, disable=not show_pbar):
        reg = ApplyWarp()
        reg.inputs.moving_image = moving
        reg.inputs.transform_file = tf
        reg.inputs.output_path = output_path
        reg.terminal_output = preferences.nipype_logging
        reg.inputs.num_threads = num_threads
        reg_output = reg.run(cwd=cwd)

        moving = reg_output.outputs.warped_file

    if rtype == MedicalVolume:
        out = NiftiReader().load(moving)
    else:
        out = moving

    if os.path.isfile(mv_filepath):
        os.remove(mv_filepath)
    if not has_output_path and os.path.isdir(output_path):
        shutil.rmtree(output_path)

    return out


def _apply_warp_mp(args, **kwargs):
    """Reorder arguments for multiprocessing support."""
    moving, output_path = args
    return _apply_warp(moving=moving, output_path=output_path, **kwargs)


def _write(vol: MedicalVolume, path: str):
    """Extracted out for multiprocessing purposes."""
    NiftiWriter().save(vol, path)


def _read(path: str):
    return NiftiReader().load(path)


def _local_exe(exe):
    """Returns path to local executable if exists, else None."""
    assert exe in ["elastix", "transformix"]
    dosma_path = os.path.join(fc._DOSMA_ELASTIX_FOLDER, exe)
    if os.path.isfile(dosma_path):
        return os.path.abspath(dosma_path)


def _local_lib_dir():
    """Returns path to directory with local lib file if exists, else None."""
    files = [x for x in os.listdir(fc._DOSMA_ELASTIX_FOLDER) if x.startswith("libANNlib")]
    if len(files) > 0:
        return fc._DOSMA_ELASTIX_FOLDER


def _is_main_process():
    py_version = tuple(sys.version_info[0:2])
    return (py_version < (3, 8) and mp.current_process().name == "MainProcess") or (
        py_version >= (3, 8) and mp.parent_process() is None
    )



"""NIfTI I/O.

This module contains NIfTI input/output helpers.


"""

import os
from typing import Collection

import nibabel as nib

from dosma.core.io.format_io import DataReader, DataWriter, ImageDataFormat
from dosma.core.med_volume import MedicalVolume
from dosma.defaults import AFFINE_DECIMAL_PRECISION, SCANNER_ORIGIN_DECIMAL_PRECISION
from dosma.utils import io_utils

__all__ = ["NiftiReader", "NiftiWriter"]


class NiftiReader(DataReader):
    """A class for reading NIfTI files.

    Attributes:
        data_format_code (ImageDataFormat): The supported image data format.
    """

    data_format_code = ImageDataFormat.nifti

    def load(self, file_path, mmap: bool = False) -> MedicalVolume:
        """Load volume from NIfTI file path.

        A NIfTI file should only correspond to one volume.

        Args:
            file_path (str): File path to NIfTI file.
            mmap (bool): Whether to use memory mapping.

        Returns:
            MedicalVolume: Loaded volume.

        Raises:
            FileNotFoundError: If `file_path` not found.
            ValueError: If `file_path` does not end in a supported NIfTI extension.
        """
        if not os.path.isfile(file_path):
            raise FileNotFoundError("{} not found".format(file_path))

        if not self.data_format_code.is_filetype(file_path):
            raise ValueError(
                "{} must be a file with extension '.nii' or '.nii.gz'".format(file_path)
            )

        nib_img = nib.load(file_path)
        return MedicalVolume.from_nib(
            nib_img,
            affine_precision=AFFINE_DECIMAL_PRECISION,
            origin_precision=SCANNER_ORIGIN_DECIMAL_PRECISION,
            mmap=mmap,
        )

    def __serializable_variables__(self) -> Collection[str]:
        return self.__dict__.keys()

    read = load  # pragma: no cover


class NiftiWriter(DataWriter):
    """A class for writing volumes in NIfTI format.

    Attributes:
        data_format_code (ImageDataFormat): The supported image data format.
    """

    data_format_code = ImageDataFormat.nifti

    def save(self, volume: MedicalVolume, file_path: str):
        """Save volume in NIfTI format,

        Args:
            volume (MedicalVolume): Volume to save.
            file_path (str): File path to NIfTI file.

        Raises:
            ValueError: If `file_path` does not end in a supported NIfTI extension.
        """
        if not self.data_format_code.is_filetype(file_path):
            raise ValueError(
                "{} must be a file with extension '.nii' or '.nii.gz'".format(file_path)
            )

        # Create dir if does not exist
        io_utils.mkdirs(os.path.dirname(file_path))

        nib_img = volume.to_nib()
        nib.save(nib_img, file_path)

    def __serializable_variables__(self) -> Collection[str]:
        return self.__dict__.keys()

    write = save  # pragma: no cover



import unittest

from dosma.models import util as m_util


class TestUtil(unittest.TestCase):
    def test_aliases_exist(self):
        # Verify that none of the supported segmentation models have overlapping aliases
        models = m_util.__SUPPORTED_MODELS__
        alias_to_model = {}  # noqa: F841

        for m in models:
            aliases = m.ALIASES

            # all supported models must have at least 1 alias that is not ''
            valid_alias = len(aliases) >= 1 and all([x != "" for x in aliases])

            assert valid_alias, "%s does not have valid aliases" % m

    def test_overlapping_aliases(self):
        # Verify that none of the supported segmentation models have overlapping aliases
        models = m_util.__SUPPORTED_MODELS__
        alias_to_model = {}

        for m in models:
            curr_aliases = set(alias_to_model.keys())
            aliases = set(m.ALIASES)

            assert aliases.isdisjoint(curr_aliases), "%s alias(es) already in use" % str(
                aliases.intersection(curr_aliases)
            )

            for a in aliases:
                alias_to_model[a] = m



__all__ = ["ActionWrapper"]


class ActionWrapper(object):
    """Wrapper for actions (methods) that can be executed via command-line.

    Examples include `segment` scans, `interregister` scans, etc.

    Actions are instance methods of classes that can be executed via the command line.
    They are typically associated with different scans.

    To expose these methods to the command-line interface, we wrap these actions as subparsers.
        Parameters for the method are arguments of the subparser.
    """

    def __init__(self, name, **kwargs):
        self._name = name
        self._help = ""
        self._param_help = None
        self._alternative_param_names = None
        self._aliases = []

        if "help" in kwargs:
            self._help = kwargs.get("help")

        if "aliases" in kwargs:
            aliases = kwargs.get("aliases")
            assert type(aliases) is list, "aliases must be a list"
            self._aliases = aliases

        if "param_help" in kwargs:
            param_help_in = kwargs.get("param_help")
            assert type(param_help_in) is dict, "`param_help` must be a dictionary of str->str"
            for param_name in param_help_in:
                assert type(param_name) is str, "Keys must be of string type"
                assert type(param_help_in[param_name]) is str, "Values must be of string type"
            self._param_help = dict(param_help_in)

        if "alternative_param_names" in kwargs:
            alternative_param_names_in = kwargs.get("alternative_param_names")
            assert (
                type(alternative_param_names_in) is dict
            ), "`alternative_param_names` must be a dictionary of str->str"
            for param_name in alternative_param_names_in:
                assert type(param_name) is str, "Keys must be of string type"
                assert type(alternative_param_names_in[param_name]) in [
                    list,
                    tuple,
                ], "Values must be of string type"
            self._alternative_param_names = alternative_param_names_in

    def get_alternative_param_names(self, param: str):
        """Get aliases (alternate names) for a parameter.

        Args:
            param (str): Action parameter name.

        Returns:
            Optional[list[str]]: If aliases exist for parameter. `None`, otherwise.
        """
        if not self._alternative_param_names or param not in self._alternative_param_names:
            return None

        return self._alternative_param_names[param]

    def get_param_help(self, param: str):
        """Get help menu for a parameter.

        Args:
            param (str): Action parameter name.

        Returns:
            str: Help menu for parameter, if initialized. `""`, otherwise.
        """
        if not self._param_help or param not in self._param_help:
            return ""

        return self._param_help[param]

    @property
    def aliases(self):
        """list[str]: Aliases (other names) for this action."""
        return self._aliases

    @property
    def help(self):
        """str: Help menu for this action."""
        return self._help

    @property
    def name(self):
        """str: Action name."""
        return self._name
